#
# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import datetime
import copy
import unittest
import pytest

from glyphsLib.classes import (
    GSFont,
    GSFontMaster,
    GSInstance,
    GSCustomParameter,
    GSGlyph,
    GSLayer,
    GSAnchor,
    GSComponent,
    GSAlignmentZone,
    GSClass,
    GSFeature,
    GSAnnotation,
    GSFeaturePrefix,
    GSGuide,
    GSHint,
    GSNode,
    GSPath,
    GSSmartComponentAxis,
    GSBackgroundImage,
    InstanceType,
    segment,
    LayerComponentsProxy,
    LayerGuideLinesProxy,
    STEM,
    TEXT,
    ARROW,
    CIRCLE,
    PLUS,
    MINUS,
    LINE,
    CURVE,
    OFFCURVE,
)
from glyphsLib.types import Point, Transform, Rect


TESTFILE_PATH = os.path.join(
    os.path.dirname(__file__), os.path.join("data", "GlyphsUnitTestSans.glyphs")
)


def generate_minimal_font(format_version=2):
    font = GSFont()
    font.format_version = format_version
    font.appVersion = 895
    font.date = datetime.datetime.today()
    font.familyName = "MyFont"

    master = GSFontMaster()
    master.ascender = 0
    master.capHeight = 0
    master.descender = 0
    master.id = "id"
    master.xHeight = 0
    font.masters.append(master)

    font.upm = 1000
    font.versionMajor = 1
    font.versionMinor = 0

    return font


def generate_instance_from_dict(instance_dict):
    instance = GSInstance()
    instance.name = instance_dict["name"]
    for custom_parameter in instance_dict.get("customParameters", []):
        cp = GSCustomParameter()
        cp.name = custom_parameter["name"]
        cp.value = custom_parameter["value"]
        instance.customParameters.append(cp)
    return instance


def add_glyph(font, glyphname):
    glyph = GSGlyph()
    glyph.name = glyphname
    font.glyphs.append(glyph)
    layer = GSLayer()
    layer.layerId = font.masters[0].id
    layer.associatedMasterId = font.masters[0].id
    layer.width = 0
    glyph.layers.append(layer)
    return glyph


def add_anchor(font, glyphname, anchorname, x, y):
    for glyph in font.glyphs:
        if glyph.name == glyphname:
            for master in font.masters:
                layer = glyph.layers[master.id]
                layer.anchors = getattr(layer, "anchors", [])
                anchor = GSAnchor()
                anchor.name = anchorname
                anchor.position = Point(x, y)
                layer.anchors.append(anchor)


def add_component(font, glyphname, componentname, transform):
    for glyph in font.glyphs:
        if glyph.name == glyphname:
            for layer in glyph.layers.values():
                component = GSComponent()
                component.name = componentname
                component.transform = transform
                layer.components.append(component)


class GlyphLayersTest(unittest.TestCase):
    def test_check_master_layer(self):
        font = generate_minimal_font()
        glyph = add_glyph(font, "A")
        self.assertIsNotNone(glyph)
        master = font.masters[0]
        self.assertIsNotNone(master)
        layer = glyph.layers[master.id]
        self.assertIsNotNone(layer)

        layer = glyph.layers["XYZ123"]
        self.assertIsNone(layer)

    def test_append_layer_same_id(self):
        font = generate_minimal_font()
        font.masters[0].id = "abc"
        master2 = GSFontMaster()
        master2.ascender = 0
        master2.capHeight = 0
        master2.descender = 0
        master2.xHeight = 0
        master2.id = "abc"
        font.masters.append(master2)
        assert len({m.id for m in font.masters}) == 2

    def test_iterate_layers_of_orphan_glyph(self):
        # https://github.com/googlefonts/glyphsLib/issues/1013
        glyph = GSGlyph()
        assert glyph.parent is None
        layer = GSLayer()
        glyph.layers.append(layer)
        assert layer.parent is glyph
        # this ought not to raise a `KeyError: 0` exception
        layers = list(glyph.layers)
        assert layers[0] is layer


class GSFontTest(unittest.TestCase):
    def test_init(self):
        font = GSFont()
        self.assertEqual(font.familyName, "Unnamed font")
        self.assertEqual(font.versionMajor, 1)
        self.assertEqual(font.versionMinor, 0)
        self.assertEqual(font.appVersion, "895")

        self.assertEqual(len(font.glyphs), 0)
        self.assertEqual(len(font.masters), 0)
        self.assertEqual(list(font.masters), list(()))
        self.assertEqual(len(font.instances), 0)
        self.assertEqual(font.instances, [])
        self.assertEqual(len(font.customParameters), 0)

    def test_repr(self):
        font = GSFont()
        self.assertEqual(repr(font), '<GSFont "Unnamed font">')

    def test_update_custom_parameter(self):
        font = GSFont()
        font.customParameters["Filter"] = "RemoveOverlap"
        self.assertEqual(font.customParameters["Filter"], "RemoveOverlap")
        font.customParameters["Filter"] = "AddExtremes"
        self.assertEqual(font.customParameters["Filter"], "AddExtremes")

    def test_font_master_proxy(self):
        font = GSFont()
        master = GSFontMaster()
        font.masters.append(master)
        self.assertEqual(master.font, font)


class GSInstanceTest(unittest.TestCase):
    def test_variable_instance(self):
        instance = GSInstance()
        instance.name = "Variable"
        instance.type = InstanceType.VARIABLE

        assert instance.weightValue is None
        assert instance.widthValue is None
        assert instance.customValue is None
        assert instance.customValue1 is None
        assert instance.customValue2 is None
        assert instance.customValue3 is None


class GSObjectsTestCase(unittest.TestCase):
    def setUp(self):
        self.font = GSFont(TESTFILE_PATH)

    def assertString(self, value):
        self.assertIsInstance(value, str)
        old_value = value
        value = "eee"
        self.assertEqual(value, "eee")
        value = old_value
        self.assertEqual(value, old_value)

    def assertUnicode(self, value):
        self.assertIsInstance(value, str)
        old_value = value
        value = "ə"
        self.assertEqual(value, "ə")
        value = old_value
        self.assertEqual(value, old_value)

    def assertInteger(self, value):
        self.assertIsInstance(value, int)
        old_value = value
        value = 5
        self.assertEqual(value, 5)
        value = old_value
        self.assertEqual(value, old_value)

    def assertFloat(self, value):
        self.assertIsInstance(value, float)
        old_value = value
        value = 0.5
        self.assertEqual(value, 0.5)
        value = old_value
        self.assertEqual(value, old_value)

    def assertBool(self, value):
        self.assertIsInstance(value, bool)
        old_value = value
        value = not value
        self.assertEqual(value, not old_value)
        value = old_value
        self.assertEqual(value, old_value)

    def assertDict(self, dictObject):
        self.assertIsInstance(dictObject, dict)
        var1 = "abc"
        var2 = "def"
        dictObject["uniTestValue"] = var1
        self.assertEqual(dictObject["uniTestValue"], var1)
        dictObject["uniTestValue"] = var2
        self.assertEqual(dictObject["uniTestValue"], var2)


class GSFontFromFileTest(GSObjectsTestCase):
    def setUp(self):
        super().setUp()

    def test_pathlike_path(self):
        from pathlib import Path

        font = GSFont(TESTFILE_PATH)
        self.assertEqual(font.filepath, TESTFILE_PATH)

        font = GSFont(Path(TESTFILE_PATH))
        self.assertEqual(font.filepath, TESTFILE_PATH)

    def test_masters(self):
        font = self.font
        amount = len(font.masters)
        self.assertEqual(len(list(font.masters)), 3)

        new_master = GSFontMaster()
        font.masters.append(new_master)
        self.assertEqual(new_master, font.masters[-1])
        del font.masters[-1]

        new_master1 = GSFontMaster()
        new_master2 = GSFontMaster()
        font.masters.extend([new_master1, new_master2])
        self.assertEqual(new_master1, font.masters[-2])
        self.assertEqual(new_master2, font.masters[-1])

        font.masters.remove(font.masters[-1])
        font.masters.remove(font.masters[-1])

        new_master = GSFontMaster()
        font.masters.insert(0, new_master)
        self.assertEqual(new_master, font.masters[0])
        font.masters.remove(font.masters[0])
        self.assertEqual(amount, len(font.masters))

    def test_instances(self):
        font = self.font
        amount = len(font.instances)
        self.assertEqual(len(list(font.instances)), 8)
        new_instance = GSInstance()
        font.instances.append(new_instance)
        self.assertEqual(new_instance, font.instances[-1])
        del font.instances[-1]
        new_instance1 = GSInstance()
        new_instance2 = GSInstance()
        font.instances.extend([new_instance1, new_instance2])
        self.assertEqual(new_instance1, font.instances[-2])
        self.assertEqual(new_instance2, font.instances[-1])
        font.instances.remove(font.instances[-1])
        font.instances.remove(font.instances[-1])
        new_instance = GSInstance()
        font.instances.insert(0, new_instance)
        self.assertEqual(new_instance, font.instances[0])
        font.instances.remove(font.instances[0])
        self.assertEqual(amount, len(font.instances))

    def test_glyphs(self):
        font = self.font
        self.assertGreaterEqual(len(list(font.glyphs)), 1)
        by_index = font.glyphs[4]
        by_name = font.glyphs["adieresis"]
        by_unicode_char = font.glyphs["ä"]
        by_unicode_value = font.glyphs["00E4"]
        by_unicode_value_lowercased = font.glyphs["00e4"]
        self.assertEqual(by_index, by_name)
        self.assertEqual(by_unicode_char, by_name)
        self.assertEqual(by_unicode_value, by_name)
        self.assertEqual(by_unicode_value_lowercased, by_name)

    def test_classes(self):
        font = self.font
        font.classes = []
        amount = len(font.classes)
        font.classes.append(GSClass("uppercaseLetters", "A"))
        self.assertIsNotNone(font.classes[-1].__repr__())
        self.assertEqual(len(font.classes), 1)
        self.assertIn('<GSClass "uppercaseLetters">', str(font.classes))
        self.assertIn("A", font.classes["uppercaseLetters"].code)
        del font.classes["uppercaseLetters"]
        newClass1 = GSClass("uppercaseLetters1", "A")
        newClass2 = GSClass("uppercaseLetters2", "A")
        font.classes.extend([newClass1, newClass2])
        self.assertEqual(newClass1, font.classes[-2])
        self.assertEqual(newClass2, font.classes[-1])
        newClass = GSClass("uppercaseLetters3", "A")
        newClass = copy.copy(newClass)
        font.classes.insert(0, newClass)
        self.assertEqual(newClass, font.classes[0])
        font.classes.remove(font.classes[-1])
        font.classes.remove(font.classes[-1])
        font.classes.remove(font.classes[0])
        self.assertEqual(len(font.classes), amount)

    def test_features(self):
        font = self.font
        font.features = []
        amount = len(font.features)
        font.features.append(GSFeature("liga", "sub f i by fi;"))
        # TODO
        # self.assertIsNotNone(font.features['liga'].__repr__())
        self.assertEqual(len(font.features), 1)
        # TODO
        # self.assertIn('<GSFeature "liga">', str(font.features))
        # self.assertIn('sub f i by fi;', font.features['liga'].code)
        # del(font.features['liga'])
        del font.features[-1]
        newFeature1 = GSFeature("liga", "sub f i by fi;")
        newFeature2 = GSFeature("liga", "sub f l by fl;")
        font.features.extend([newFeature1, newFeature2])
        self.assertEqual(newFeature1, font.features[-2])
        self.assertEqual(newFeature2, font.features[-1])
        newFeature = GSFeature("liga", "sub f i by fi;")
        newFeature = copy.copy(newFeature)
        font.features.insert(0, newFeature)
        self.assertEqual(newFeature, font.features[0])
        font.features.remove(font.features[-1])
        font.features.remove(font.features[-1])
        font.features.remove(font.features[0])
        self.assertEqual(len(font.features), amount)

    def test_featurePrefixes(self):
        font = self.font
        font.featurePrefixes = []
        amount = len(font.featurePrefixes)
        font.featurePrefixes.append(
            GSFeaturePrefix("LanguageSystems", "languagesystem DFLT dflt;")
        )
        self.assertIsNotNone(font.featurePrefixes[-1].__repr__())
        self.assertEqual(len(font.featurePrefixes), 1)
        self.assertIn('<GSFeaturePrefix "LanguageSystems">', str(font.featurePrefixes))
        self.assertIn("languagesystem DFLT dflt;", font.featurePrefixes[-1].code)
        # TODO
        # del(font.featurePrefixes['LanguageSystems'])
        del font.featurePrefixes[-1]
        newFeaturePrefix1 = GSFeaturePrefix(
            "LanguageSystems1", "languagesystem DFLT dflt;"
        )
        newFeaturePrefix2 = GSFeaturePrefix(
            "LanguageSystems2", "languagesystem DFLT dflt;"
        )
        font.featurePrefixes.extend([newFeaturePrefix1, newFeaturePrefix2])
        self.assertEqual(newFeaturePrefix1, font.featurePrefixes[-2])
        self.assertEqual(newFeaturePrefix2, font.featurePrefixes[-1])
        newFeaturePrefix = GSFeaturePrefix(
            "LanguageSystems3", "languagesystem DFLT dflt;"
        )
        newFeaturePrefix = copy.copy(newFeaturePrefix)
        font.featurePrefixes.insert(0, newFeaturePrefix)
        self.assertEqual(newFeaturePrefix, font.featurePrefixes[0])
        font.featurePrefixes.remove(font.featurePrefixes[-1])
        font.featurePrefixes.remove(font.featurePrefixes[-1])
        font.featurePrefixes.remove(font.featurePrefixes[0])
        self.assertEqual(len(font.featurePrefixes), amount)

    def test_ints(self):
        attributes = ["versionMajor", "versionMajor", "upm", "grid", "gridSubDivisions"]
        font = self.font
        for attr in attributes:
            self.assertInteger(getattr(font, attr))

    def test_strings(self):
        attributes = [
            "copyright",
            "designer",
            "designerURL",
            "manufacturer",
            "manufacturerURL",
            "familyName",
        ]
        font = self.font
        for attr in attributes:
            self.assertUnicode(getattr(font, attr))

    def test_note(self):
        font = self.font
        self.assertUnicode(font.note)

    # date
    def test_date(self):
        font = self.font
        self.assertIsInstance(font.date, datetime.datetime)

    def test_kerning(self):
        font = self.font
        self.assertDict(font.kerning)

    def test_userData(self):
        font = self.font
        self.assertEqual(
            font.userData["AsteriskParameters"],
            {"253E7231-480D-4F8E-8754-50FC8575C08E": ["754", "30", 7, 51, "80", "50"]},
        )
        # self.assertIsInstance(font.userData, dict)
        # TODO
        self.assertIsNone(font.userData["TestData"])
        font.userData["TestData"] = 42
        self.assertEqual(font.userData["TestData"], 42)
        self.assertTrue("TestData" in font.userData)
        del font.userData["TestData"]
        self.assertIsNone(font.userData["TestData"])

    def test_disableNiceNames(self):
        font = self.font
        self.assertIsInstance(font.disablesNiceNames, bool)

    def test_customParameters(self):
        font = self.font

        # disabled custom parameters are ignored by getters-by-name
        assert font.customParameters["Axis Mappings"] is None
        assert font.customParameters.get("Axis Mappings") is None
        assert "Axis Mappings" not in font.customParameters
        # ... however this test font actually contains one
        cp = font.customParameters._get_by_name("Axis Mappings", skip_disabled=False)
        assert cp.disabled
        assert cp.value == {"wght": {}}
        # On the other hand, __iter__, __len__ or get-by-index still take into account
        # the presence of the disabled parameter. Yeah it's weird, but Glyphs.app
        # also does the same...
        assert cp in list(font.customParameters)
        assert font.customParameters[0] == cp
        assert len(font.customParameters) == 2

        assert font.customParameters[1].name == "note"
        assert font.customParameters[1].value == "Bla bla"

        font.customParameters["trademark"] = "ThisFont is a trademark by MyFoundry.com"
        self.assertIn(
            font.customParameters["trademark"],
            "ThisFont is a trademark by MyFoundry.com",
        )
        amount = len(list(font.customParameters))
        newParameter = GSCustomParameter("hello1", "world1")
        font.customParameters.append(newParameter)
        self.assertEqual(newParameter, list(font.customParameters)[-1])
        del font.customParameters[-1]
        newParameter1 = GSCustomParameter("hello2", "world2")
        newParameter2 = GSCustomParameter("hello3", "world3")
        newParameter2 = copy.copy(newParameter2)
        font.customParameters.extend([newParameter1, newParameter2])
        self.assertEqual(newParameter1, list(font.customParameters)[-2])
        self.assertEqual(newParameter2, list(font.customParameters)[-1])
        font.customParameters.remove(list(font.customParameters)[-1])
        font.customParameters.remove(list(font.customParameters)[-1])
        newParameter = GSCustomParameter("hello1", "world1")
        font.customParameters.insert(0, newParameter)
        self.assertEqual(newParameter, list(font.customParameters)[0])
        font.customParameters.remove(list(font.customParameters)[0])
        self.assertEqual(amount, len(list(font.customParameters)))
        del font.customParameters["trademark"]

    # TODO: selection, selectedLayers, currentText, tabs, currentTab

    # TODO: selectedFontMaster, masterIndex

    def test_filepath(self):
        font = self.font
        self.assertIsNotNone(font.filepath)

    # TODO: tool, tools
    # TODO: save(), close()
    # TODO: setKerningForPair(), kerningForPair(), removeKerningForPair()
    # TODO: updateFeatures()
    # TODO: copy(font)


class GSFontMasterFromFileTest(GSObjectsTestCase):
    def setUp(self):
        super().setUp()
        self.font = GSFont(TESTFILE_PATH)
        self.master = self.font.masters[0]

    def test_attributes(self):
        master = self.master
        self.assertIsNotNone(master.__repr__())
        self.assertIsNotNone(master.id)
        self.assertIsNotNone(master.name)
        self.assertIsNotNone(master.weight)
        self.assertIsNotNone(master.width)
        # weightValue
        obj = master.weightValue
        old_obj = obj
        self.assertIsInstance(obj, int)
        master.weightValue = 0.5
        self.assertEqual(master.weightValue, 0.5)
        self.assertIsInstance(master.weightValue, float)
        master.weightValue = old_obj
        self.assertIsInstance(master.widthValue, int)
        self.assertIsInstance(master.customValue, int)
        self.assertIsInstance(master.ascender, int)
        self.assertIsInstance(master.capHeight, int)
        self.assertIsInstance(master.xHeight, int)
        self.assertIsInstance(master.descender, int)
        self.assertIsInstance(master.italicAngle, int)
        for attr in [
            "weightValue",
            "widthValue",
            "customValue",
            "ascender",
            "capHeight",
            "xHeight",
            "descender",
            "italicAngle",
        ]:
            value = getattr(master, attr)
            self.assertIsInstance(value, int)
            setattr(master, attr, 0.5)
            self.assertEqual(getattr(master, attr), 0.5)
            setattr(master, attr, value)
        self.assertIsInstance(master.customName, str)

        # verticalStems
        oldStems = master.verticalStems
        master.verticalStems = [10, 15, 20]
        self.assertEqual(len(master.verticalStems), 3)
        master.verticalStems = oldStems

        # horizontalStems
        oldStems = master.horizontalStems
        master.horizontalStems = [10, 15, 20]
        self.assertEqual(len(master.horizontalStems), 3)
        master.horizontalStems = oldStems

        # alignmentZones
        self.assertIsInstance(master.alignmentZones, list)

        # TODO blueValues
        # self.assertIsInstance(master.blueValues, list)

        # TODO otherBlues
        # self.assertIsInstance(master.otherBlues, list)

        # guides
        self.assertIsInstance(master.guides, list)
        master.guides = []
        self.assertEqual(len(master.guides), 0)
        newGuide = GSGuide()
        newGuide.position = Point("{100, 100}")
        newGuide.angle = -10.0
        master.guides.append(newGuide)
        self.assertIsNotNone(master.guides[0].__repr__())
        self.assertEqual(len(master.guides), 1)
        del master.guides[0]
        self.assertEqual(len(master.guides), 0)

        # guides
        self.assertIsInstance(master.guides, list)
        master.guides = []
        self.assertEqual(len(master.guides), 0)
        newGuide = GSGuide()
        newGuide.position = Point("{100, 100}")
        newGuide.angle = -10.0
        master.guides.append(newGuide)
        self.assertIsNotNone(master.guides[0].__repr__())
        self.assertEqual(len(master.guides), 1)
        del master.guides[0]
        self.assertEqual(len(master.guides), 0)

        # userData
        self.assertIsNotNone(master.userData)
        master.userData["TestData"] = 42
        self.assertEqual(master.userData["TestData"], 42)
        del master.userData["TestData"]
        # TODO
        # self.assertIsNone(master.userData["TestData"])

        # customParameters
        master.customParameters["trademark"] = (
            "ThisFont is a trademark by MyFoundry.com"
        )
        self.assertGreaterEqual(len(master.customParameters), 1)
        del master.customParameters["trademark"]

        # font
        self.assertEqual(self.font, self.master.font)

    def test_name(self):
        master = self.master
        self.assertEqual("Light", master.name)

        master.width = "Condensed"
        self.assertEqual(master.name, "Condensed Light")
        master.width = ""

        master.customParameters["Master Name"] = "My custom master name"
        self.assertEqual("My custom master name", master.name)
        del master.customParameters["Master Name"]
        self.assertEqual("Light", master.name)

        master.italicAngle = 11
        self.assertEqual("Light Italic", master.name)
        master.italicAngle = 0

        master.italicAngle = 11
        master.width = "Condensed"
        self.assertEqual("Condensed Light Italic", master.name)
        master.width = ""
        master.italicAngle = 0

        master.customName = "Rounded"
        self.assertEqual("Light Rounded", master.name)
        master.customName = "Rounded Stretched Filled Rotated"
        self.assertEqual("Light Rounded Stretched Filled Rotated", master.name)
        master.customName = ""
        self.assertEqual("Light", master.name)

        # Test the name of a master set to "Regular" in the UI dropdown
        # but with a customName
        thin = GSFontMaster()
        thin.customName = "Thin"
        self.assertEqual("Thin", thin.name)

        # Test that we don't get an extra "Regular" in the name of "Italic"
        # https://github.com/googlefonts/glyphsLib/issues/380
        master = GSFontMaster()
        master.weight = "Regular"
        master.width = "Regular"
        master.italicAngle = 10.0
        self.assertEqual("Italic", master.name)

        # Test that bold italic gets properly named.
        master = GSFontMaster()
        master.weight = "Bold"
        master.width = "Regular"
        master.italicAngle = 10.0
        self.assertEqual("Bold Italic", master.name)

        # Test that we don't get an extra "Italic" in the name of masters
        # whose customName already contain the string "Italic"
        master = GSFontMaster()
        master.weight = "Regular"
        master.width = "Regular"
        master.customName = "Italic"
        master.italicAngle = 10.0
        self.assertEqual("Italic", master.name)

        # Test that we don't get an extra "Italic" in the name of masters
        # whose customName contains the string "Oblique"
        master = GSFontMaster()
        master.weight = "Regular"
        master.width = "Regular"
        master.customName = "Oblique"
        master.italicAngle = 10.0
        self.assertEqual("Oblique", master.name)

    def test_name_assignment(self):
        test_data = [
            # <name>, <expected weight>, <expected width>, <expected custom>
            # Regular
            ("Regular", "", "", ""),
            # Weights from the dropdown
            ("Light", "Light", "", ""),
            ("SemiLight", "SemiLight", "", ""),
            ("SemiBold", "SemiBold", "", ""),
            ("Bold", "Bold", "", ""),
            # Widths from the dropdown
            ("Condensed", "", "Condensed", ""),
            ("SemiCondensed", "", "SemiCondensed", ""),
            ("SemiExtended", "", "SemiExtended", ""),
            ("Extended", "", "Extended", ""),
            # Mixed weight and width from dropdowns
            ("Light Condensed", "Light", "Condensed", ""),
            ("Bold SemiExtended", "Bold", "SemiExtended", ""),
            # With italic -- in Glyphs 1114  works like a custom part
            ("Light Italic", "Light", "", "Italic"),
            ("SemiLight Italic", "SemiLight", "", "Italic"),
            ("SemiBold Italic", "SemiBold", "", "Italic"),
            ("Bold Italic", "Bold", "", "Italic"),
            ("Condensed Italic", "", "Condensed", "Italic"),
            ("SemiCondensed Italic", "", "SemiCondensed", "Italic"),
            ("SemiExtended Italic", "", "SemiExtended", "Italic"),
            ("Extended Italic", "", "Extended", "Italic"),
            ("Light Condensed Italic", "Light", "Condensed", "Italic"),
            ("Bold SemiExtended Italic", "Bold", "SemiExtended", "Italic"),
            # With custom parts
            ("Thin", "", "", "Thin"),
            ("SemiLight Ultra Expanded", "SemiLight", "", "Ultra Expanded"),
            ("Bold Compressed", "Bold", "", "Compressed"),
            ("Fat Condensed", "", "Condensed", "Fat"),
            ("Ultra Light Extended", "Light", "Extended", "Ultra"),
            ("Hyper Light Condensed Dunhill", "Light", "Condensed", "Hyper  Dunhill"),
            ("Bold SemiExtended Rugged", "Bold", "SemiExtended", "Rugged"),
            ("Thin Italic", "", "", "Thin Italic"),
            (
                "SemiLight Ultra Expanded Italic",
                "SemiLight",
                "",
                "Ultra Expanded Italic",
            ),
            ("Bold Compressed Italic", "Bold", "", "Compressed Italic"),
            ("Fat Condensed Italic", "", "Condensed", "Fat Italic"),
            ("Ultra Light Extended Italic", "Light", "Extended", "Ultra  Italic"),
            (
                "Hyper Light Condensed Dunhill Italic",
                "Light",
                "Condensed",
                "Hyper  Dunhill Italic",
            ),
            (
                "Bold SemiExtended Rugged Italic",
                "Bold",
                "SemiExtended",
                "Rugged Italic",
            ),
            ("Green Light Red Extended Blue", "Light", "Extended", "Green Red Blue"),
            (
                "Green SemiExtended Red SemiBold Blue",
                "SemiBold",
                "SemiExtended",
                "Green Red Blue",
            ),
        ]
        master = GSFontMaster()
        for name, weight, width, custom in test_data:
            master.name = name
            self.assertEqual(master.name, name)
            self.assertEqual(master.weight, weight or "Regular")
            self.assertEqual(master.width, width or "Regular")
            self.assertEqual(master.customName, custom)

    def test_default_values(self):
        master = GSFontMaster()
        self.assertEqual(master.weightValue, 100.0)
        self.assertEqual(master.widthValue, 100.0)
        self.assertEqual(master.customValue, 0.0)
        self.assertEqual(master.customValue1, 0.0)
        self.assertEqual(master.customValue2, 0.0)
        self.assertEqual(master.customValue3, 0.0)


class GSAlignmentZoneFromFileTest(GSObjectsTestCase):
    def setUp(self):
        super().setUp()
        self.master = self.font.masters[0]

    def test_attributes(self):
        master = self.master
        for i, zone in enumerate(
            [(800, 10), (700, 10), (470, 10), (0, -10), (-200, -10)]
        ):
            pos, size = zone
            self.assertEqual(master.alignmentZones[i].position, pos)
            self.assertEqual(master.alignmentZones[i].size, size)
        master.alignmentZones = []
        self.assertEqual(len(master.alignmentZones), 0)
        master.alignmentZones.append(GSAlignmentZone(100, 10))
        self.assertIsNotNone(master.alignmentZones[-1].__repr__())
        zone = copy.copy(master.alignmentZones[-1])
        self.assertEqual(len(master.alignmentZones), 1)
        self.assertEqual(master.alignmentZones[-1].position, 100)
        self.assertEqual(master.alignmentZones[-1].size, 10)
        del master.alignmentZones[-1]
        self.assertEqual(len(master.alignmentZones), 0)


class GSInstanceFromFileTest(GSObjectsTestCase):
    def setUp(self):
        super().setUp()
        self.instance = self.font.instances[0]

    def test_attributes(self):
        instance = self.instance
        self.assertIsNotNone(instance.__repr__())

        # TODO: active
        # self.assertIsInstance(instance.active, bool)

        # name
        self.assertIsInstance(instance.name, str)

        # weight
        self.assertIsInstance(instance.weight, str)

        # width
        self.assertIsInstance(instance.width, str)

        # weightValue
        # widthValue
        # customValue
        for attr in ["weightValue", "widthValue", "customValue"]:
            value = getattr(instance, attr)
            self.assertIsInstance(value, int)
            setattr(instance, attr, 0.5)
            self.assertEqual(getattr(instance, attr), 0.5)
            setattr(instance, attr, value)
        # isItalic
        # isBold
        for attr in ["isItalic", "isBold"]:
            value = getattr(instance, attr)
            self.assertIsInstance(value, bool)
            setattr(instance, attr, not value)
            self.assertEqual(getattr(instance, attr), not value)
            setattr(instance, attr, value)

        # linkStyle
        self.assertIsInstance(instance.linkStyle, str)

        # familyName
        # preferredFamily
        # preferredSubfamilyName
        # windowsFamily
        # windowsStyle
        # windowsLinkedToStyle
        # fontName
        # fullName
        for attr in [
            "familyName",
            "preferredFamily",
            "preferredSubfamilyName",
            "windowsFamily",
            "windowsStyle",
            "windowsLinkedToStyle",
            "fontName",
            "fullName",
        ]:
            # self.assertIsInstance(getattr(instance, attr), str)
            if not hasattr(instance, attr):
                print("instance does not have %s" % attr)
                if hasattr(instance, "parent") and hasattr(instance.parent, attr):
                    value = getattr(instance.parent)
                    print(value, type(value))

        # customParameters
        instance.customParameters["trademark"] = (
            "ThisFont is a trademark by MyFoundry.com"
        )
        self.assertGreaterEqual(len(instance.customParameters), 1)
        del instance.customParameters["trademark"]

        # instanceInterpolations
        self.assertIsInstance(dict(instance.instanceInterpolations), dict)

        # manualInterpolation
        self.assertIsInstance(instance.manualInterpolation, bool)
        value = instance.manualInterpolation
        instance.manualInterpolation = not instance.manualInterpolation
        self.assertEqual(instance.manualInterpolation, not value)
        instance.manualInterpolation = value

        # interpolatedFont
        # TODO
        # self.assertIsInstance(instance.interpolatedFont, type(Glyphs.font))

        # TODO generate()

    def test_default_values(self):
        instance = GSInstance()
        self.assertEqual(instance.weightValue, 100.0)
        self.assertEqual(instance.widthValue, 100.0)
        self.assertEqual(instance.customValue, 0.0)
        self.assertEqual(instance.customValue1, 0.0)
        self.assertEqual(instance.customValue2, 0.0)
        self.assertEqual(instance.customValue3, 0.0)


class GSGlyphFromFileTest(GSObjectsTestCase):
    def setUp(self):
        super().setUp()
        self.glyph = self.font.glyphs["a"]

    # TODO duplicate
    # def test_duplicate(self):
    #     font = self.font
    #     glyph1 = self.glyph
    #     glyph2 = glyph1.duplicate()
    #     glyph3 = glyph1.duplicate('a.test')

    def test_parent(self):
        font = self.font
        glyph = self.glyph
        self.assertEqual(glyph.parent, font)

    def test_layers(self):
        glyph = self.glyph
        self.assertIsNotNone(glyph.layers)
        amount = len(glyph.layers)
        newLayer = GSLayer()
        newLayer.name = "1"
        glyph.layers.append(newLayer)
        self.assertIn('<GSLayer "1" (a)>', str(glyph.layers[-1]))
        self.assertEqual(newLayer, glyph.layers[-1])
        del glyph.layers[-1]
        newLayer1 = GSLayer()
        newLayer1.name = "2"
        newLayer2 = GSLayer()
        newLayer2.name = "3"
        glyph.layers.extend([newLayer1, newLayer2])
        self.assertEqual(newLayer1, glyph.layers[-2])
        self.assertEqual(newLayer2, glyph.layers[-1])
        newLayer = GSLayer()
        newLayer.name = "4"
        # indices here don't make sense because layer get appended using a UUID
        glyph.layers.insert(0, newLayer)
        # so the latest layer got appended at the end also
        self.assertEqual(newLayer, glyph.layers[-1])
        glyph.layers.remove(glyph.layers[-1])
        glyph.layers.remove(glyph.layers[-1])
        glyph.layers.remove(glyph.layers[-1])
        self.assertEqual(amount, len(glyph.layers))
        self.assertEqual(
            '[<GSLayer "Light" (a)>, <GSLayer "Regular" (a)>, '
            '<GSLayer "Bold" (a)>, <GSLayer "{155, 100}" (a)>]',
            repr(list(glyph.layers)),
        )
        self.assertEqual(
            '[<GSLayer "Bold" (a)>, <GSLayer "Regular" (a)>, '
            '<GSLayer "Light" (a)>, <GSLayer "{155, 100}" (a)>]',
            repr(list(glyph.layers.values())),
        )

    def test_layers_missing_master(self):
        """
        font.glyph['a'] has its layers in a different order
        than the font.masters and an extra layer.
        Adding a master but not adding it as a layer to the glyph should not
        affect glyph.layers unexpectedly.
        """
        glyph = self.glyph
        num_layers = len(glyph.layers)
        self.assertEqual(
            {l.layerId for l in glyph.layers},
            {l.layerId for l in glyph.layers.values()},
        )
        self.assertNotEqual(
            [l.layerId for l in glyph.layers],
            [l.layerId for l in glyph.layers.values()],
        )
        new_fontMaster = GSFontMaster()
        self.font.masters.insert(0, new_fontMaster)
        self.assertEqual(num_layers, len(glyph.layers))
        self.assertEqual(
            {l.layerId for l in glyph.layers},
            {l.layerId for l in glyph.layers.values()},
        )
        self.assertNotEqual(
            [l.layerId for l in glyph.layers],
            [l.layerId for l in glyph.layers.values()],
        )

    def test_name(self):
        glyph = self.glyph
        self.assertIsInstance(glyph.name, str)
        value = glyph.name
        glyph.name = "Ə"
        self.assertEqual(glyph.name, "Ə")
        glyph.name = value

    def test_unicode(self):
        glyph = self.glyph
        self.assertIsInstance(glyph.unicode, str)
        value = glyph.unicode
        # TODO:
        # glyph.unicode = "004a"
        # self.assertEqual(glyph.unicode, "004A")
        glyph.unicode = "004B"
        self.assertEqual(glyph.unicode, "004B")
        glyph.unicode = value

    def test_string(self):
        glyph = self.font.glyphs["adieresis"]
        self.assertEqual(glyph.string, "ä")

    def test_id(self):
        # TODO
        pass

    # TODO
    # category
    # storeCategory
    # subCategory
    # storeSubCategory
    # script
    # storeScript
    # productionName
    # storeProductionName
    # glyphInfo

    def test_horiz_kerningGroup(self):
        for group in ["leftKerningGroup", "rightKerningGroup"]:
            glyph = self.glyph
            self.assertIsInstance(getattr(glyph, group), str)
            value = getattr(glyph, group)
            setattr(glyph, group, "ä")
            self.assertEqual(getattr(glyph, group), "ä")
            setattr(glyph, group, value)

    def test_horiz_metricsKey(self):
        for group in ["leftMetricsKey", "rightMetricsKey"]:
            glyph = self.glyph
            if getattr(glyph, group) is not None:
                self.assertIsInstance(getattr(glyph, group), str)
            value = getattr(glyph, group)
            setattr(glyph, group, "ä")
            self.assertEqual(getattr(glyph, group), "ä")
            setattr(glyph, group, value)

    def test_export(self):
        glyph = self.glyph
        self.assertIsInstance(glyph.export, bool)
        value = glyph.export
        glyph.export = not glyph.export
        self.assertEqual(glyph.export, not value)
        glyph.export = value

    def test_color(self):
        glyph = self.glyph
        if glyph.color is not None:
            self.assertIsInstance(glyph.color, int)
        value = glyph.color
        glyph.color = 5
        self.assertEqual(glyph.color, 5)
        glyph.color = value

    def test_note(self):
        glyph = self.glyph
        if glyph.note is not None:
            self.assertIsInstance(glyph.note, str)
        value = glyph.note
        glyph.note = "ä"
        self.assertEqual(glyph.note, "ä")
        glyph.note = value

    # TODO
    # masterCompatible

    def test_userData(self):
        glyph = self.glyph
        # self.assertIsNone(glyph.userData)
        amount = len(glyph.userData)
        var1 = "abc"
        var2 = "def"
        glyph.userData["unitTestValue"] = var1
        self.assertEqual(glyph.userData["unitTestValue"], var1)
        glyph.userData["unitTestValue"] = var2
        self.assertEqual(glyph.userData["unitTestValue"], var2)
        del glyph.userData["unitTestValue"]
        self.assertIsNone(glyph.userData.get("unitTestValue"))
        self.assertEqual(len(glyph.userData), amount)

    def test_smart_component_axes(self):
        shoulder = self.font.glyphs["_part.shoulder"]
        axes = shoulder.smartComponentAxes
        self.assertIsNotNone(axes)
        crotch_depth, shoulder_width = axes
        self.assertIsInstance(crotch_depth, GSSmartComponentAxis)
        self.assertEqual("crotchDepth", crotch_depth.name)
        self.assertEqual(0, crotch_depth.topValue)
        self.assertEqual(-100, crotch_depth.bottomValue)
        self.assertIsInstance(shoulder_width, GSSmartComponentAxis)
        self.assertEqual("shoulderWidth", shoulder_width.name)
        self.assertEqual(100, shoulder_width.topValue)
        self.assertEqual(0, shoulder_width.bottomValue)

    # TODO
    # lastChange


class GSLayerFromFileTest(GSObjectsTestCase):
    def setUp(self):
        super().setUp()
        self.glyph = self.font.glyphs["a"]
        self.layer = self.glyph.layers[0]

    def test_repr(self):
        layer = self.layer
        self.assertIsNotNone(layer.__repr__())

    def test_repr_orphan_glyph(self):
        # https://github.com/googlefonts/glyphsLib/issues/1014
        layer = GSLayer()
        self.assertIsNone(layer.parent)  # orphan layer

        expected = '<GSLayer "" (orphan)>'
        self.assertEqual(repr(layer), expected)

        layer.layerId = layer.associatedMasterId = "layer-0"
        self.assertTrue(layer._is_master_layer)
        self.assertEqual(repr(layer), expected)

        parent = GSGlyph()
        parent.layers.append(layer)
        self.assertEqual(layer.parent, parent)  # no longer orphan layer
        self.assertIsNone(parent.parent)  # but still orphan glyph

        # this should not crash with
        #   AttributeError: 'NoneType' object has no attribute 'masterForId'
        self.assertEqual(repr(layer), expected)

    def test_parent(self):
        self.assertIs(self.layer.parent, self.glyph)
        self.assertIs(self.layer._background.parent, self.glyph)

    def test_name(self):
        layer = self.layer
        self.assertUnicode(layer.name)

    # TODO
    # def test_associatedMasterId(self):
    #     font = self.font
    #     layer = self.layer
    #     self.assertEqual(layer.associatedMasterId, font.masters[0].id)

    # def test_layerId(self):
    #     font = self.font
    #     layer = self.layer
    #     self.assertEqual(layer.layerId, font.masters[0].id)

    # TODO set layer color in .glyphs file
    # def test_color(self):
    #     layer = self.layer
    #     self.assertInteger(layer.color)

    def test_components(self):
        glyph = self.font.glyphs["adieresis"]
        layer = glyph.layers[0]
        self.assertIsNotNone(layer.components)
        self.assertIsInstance(layer.components, LayerComponentsProxy)
        # self.assertGreaterEqual(len(layer.components), 1)
        self.assertEqual(len(layer.components), 2)
        # for component in layer.components:
        #     self.assertIsInstance(component, GSComponent)
        #     self.assertEqual(component.parent, layer)
        amount = len(layer.components)
        component = GSComponent()
        component.name = "A"
        layer.components.append(component)
        self.assertEqual(component.parent, layer)
        self.assertEqual(len(layer.components), amount + 1)
        del layer.components[-1]
        self.assertEqual(len(layer.components), amount)
        layer.components.extend([component])
        self.assertEqual(len(layer.components), amount + 1)
        layer.components.remove(component)
        self.assertEqual(len(layer.components), amount)

    def test_guides(self):
        layer = self.layer
        self.assertIsInstance(layer.guides, LayerGuideLinesProxy)
        for guide in layer.guides:
            self.assertEqual(guide.parent, layer)
        layer.guides = []
        self.assertEqual(len(layer.guides), 0)
        newGuide = GSGuide()
        newGuide.position = Point("{100, 100}")
        newGuide.angle = -10.0
        amount = len(layer.guides)
        layer.guides.append(newGuide)
        self.assertEqual(newGuide.parent, layer)
        self.assertIsNotNone(layer.guides[0].__repr__())
        self.assertEqual(len(layer.guides), amount + 1)
        del layer.guides[0]
        self.assertEqual(len(layer.guides), amount)

    def test_annotations(self):
        layer = self.layer
        # self.assertEqual(layer.annotations, [])
        self.assertEqual(len(layer.annotations), 0)
        newAnnotation = GSAnnotation()
        newAnnotation.type = TEXT
        newAnnotation.text = "This curve is ugly!"
        layer.annotations.append(newAnnotation)
        # TODO position.x, position.y
        # self.assertIsNotNone(layer.annotations[0].__repr__())
        self.assertEqual(len(layer.annotations), 1)
        del layer.annotations[0]
        self.assertEqual(len(layer.annotations), 0)
        newAnnotation1 = GSAnnotation()
        newAnnotation1.type = ARROW
        newAnnotation2 = GSAnnotation()
        newAnnotation2.type = CIRCLE
        newAnnotation3 = GSAnnotation()
        newAnnotation3.type = PLUS
        layer.annotations.extend([newAnnotation1, newAnnotation2, newAnnotation3])
        self.assertEqual(layer.annotations[-3], newAnnotation1)
        self.assertEqual(layer.annotations[-2], newAnnotation2)
        self.assertEqual(layer.annotations[-1], newAnnotation3)
        newAnnotation = GSAnnotation()
        newAnnotation = copy.copy(newAnnotation)
        newAnnotation.type = MINUS
        layer.annotations.insert(0, newAnnotation)
        self.assertEqual(layer.annotations[0], newAnnotation)
        layer.annotations.remove(layer.annotations[0])
        layer.annotations.remove(layer.annotations[-1])
        layer.annotations.remove(layer.annotations[-1])
        layer.annotations.remove(layer.annotations[-1])
        self.assertEqual(len(layer.annotations), 0)

    def test_hints_from_file(self):
        glyph = self.font.glyphs["A"]
        layer = glyph.layers[1]
        self.assertEqual(2, len(layer.hints))
        first, second = layer.hints
        self.assertIsInstance(first, GSHint)
        self.assertTrue(first.horizontal)
        self.assertIsInstance(first.originNode, GSNode)
        first_origin_node = layer.paths[1].nodes[1]
        self.assertEqual(first_origin_node, first.originNode)

        self.assertIsInstance(second, GSHint)
        second_target_node = layer.paths[0].nodes[4]
        self.assertEqual(second_target_node, second.targetNode)

    def test_hints(self):
        layer = self.layer
        # layer.hints = []
        self.assertEqual(len(layer.hints), 0)
        newHint = GSHint()
        newHint = copy.copy(newHint)
        newHint.originNode = layer.paths[0].nodes[0]
        newHint.targetNode = layer.paths[0].nodes[1]
        newHint.type = STEM
        layer.hints.append(newHint)
        self.assertIsNotNone(layer.hints[0].__repr__())
        self.assertEqual(len(layer.hints), 1)
        del layer.hints[0]
        self.assertEqual(len(layer.hints), 0)
        newHint1 = GSHint()
        newHint1.originNode = layer.paths[0].nodes[0]
        newHint1.targetNode = layer.paths[0].nodes[1]
        newHint1.type = STEM
        newHint2 = GSHint()
        newHint2.originNode = layer.paths[0].nodes[0]
        newHint2.targetNode = layer.paths[0].nodes[1]
        newHint2.type = STEM
        layer.hints.extend([newHint1, newHint2])
        newHint = GSHint()
        newHint.originNode = layer.paths[0].nodes[0]
        newHint.targetNode = layer.paths[0].nodes[1]
        self.assertEqual(layer.hints[-2], newHint1)
        self.assertEqual(layer.hints[-1], newHint2)
        layer.hints.insert(0, newHint)
        self.assertEqual(layer.hints[0], newHint)
        layer.hints.remove(layer.hints[0])
        layer.hints.remove(layer.hints[-1])
        layer.hints.remove(layer.hints[-1])
        self.assertEqual(len(layer.hints), 0)

    def test_anchors(self):
        layer = self.layer
        amount = len(layer.anchors)
        self.assertEqual(len(layer.anchors), 3)
        for anchor in layer.anchors:
            self.assertEqual(anchor.parent, layer)
        if layer.anchors["top"]:
            oldPosition = layer.anchors["top"].position
        else:
            oldPosition = None
        layer.anchors["top"] = GSAnchor()
        self.assertGreaterEqual(len(layer.anchors), 1)
        self.assertIsNotNone(layer.anchors["top"].__repr__())
        layer.anchors["top"].position = Point("{100, 100}")
        # anchor = copy.copy(layer.anchors['top'])
        del layer.anchors["top"]
        layer.anchors["top"] = GSAnchor()
        self.assertEqual(amount, len(layer.anchors))
        layer.anchors["top"].position = oldPosition
        self.assertUnicode(layer.anchors["top"].name)
        newAnchor1 = GSAnchor()
        newAnchor1.name = "testPosition1"
        newAnchor2 = GSAnchor()
        newAnchor2.name = "testPosition2"
        layer.anchors.extend([newAnchor1, newAnchor2])
        self.assertEqual(layer.anchors["testPosition1"], newAnchor1)
        self.assertEqual(layer.anchors["testPosition2"], newAnchor2)
        newAnchor3 = GSAnchor()
        newAnchor3.name = "testPosition3"
        layer.anchors.append(newAnchor3)
        self.assertEqual(layer.anchors["testPosition3"], newAnchor3)
        layer.anchors.remove(layer.anchors["testPosition3"])
        layer.anchors.remove(layer.anchors["testPosition2"])
        layer.anchors.remove(layer.anchors["testPosition1"])
        self.assertEqual(amount, len(layer.anchors))

    # TODO layer.paths

    # TODO
    # selection

    # TODO
    # LSB, RSB, TSB, BSB, width

    def test_leftMetricsKey(self):
        self.assertIs(self.layer.leftMetricsKey, None)

    def test_rightMetricsKey(self):
        self.assertIs(self.layer.rightMetricsKey, None)

    def test_widthMetricsKey(self):
        self.assertIs(self.layer.widthMetricsKey, None)

    # TODO: bounds, selectionBounds

    def test_background(self):
        self.assertIn("GSBackgroundLayer", self.layer.background.__repr__())
        self.assertIs(self.layer, self.layer.background.foreground)
        self.assertIs(self.layer.parent, self.layer.background.parent)

    def test_backgroundImage(self):
        # The selected layer (0 of glyph 'a') doesn't have one
        self.assertIsNone(self.layer.backgroundImage)

        glyph = self.font.glyphs["A"]
        layer = glyph.layers[0]
        image = layer.backgroundImage
        self.assertIsInstance(image, GSBackgroundImage)
        # Values from the file
        self.assertEqual("A.jpg", image.path)
        self.assertEqual([0.0, 0.0, 489.0, 637.0], list(image.crop))
        # Default values
        self.assertEqual(50, image.alpha)
        self.assertEqual([1, 0, 0, 1, 0, 0], image.transform.value)
        self.assertEqual(False, image.locked)

        # Test documented behaviour of "alpha"
        image.alpha = 10
        self.assertEqual(10, image.alpha)
        image.alpha = 9
        self.assertEqual(50, image.alpha)
        image.alpha = 100
        self.assertEqual(100, image.alpha)
        image.alpha = 101
        self.assertEqual(50, image.alpha)

    # TODO?
    # bezierPath, openBezierPath, completeBezierPath, completeOpenBezierPath?

    def test_userData(self):
        layer = self.layer
        # self.assertDict(layer.userData)
        layer.userData["Hallo"] = "Welt"
        self.assertEqual(layer.userData["Hallo"], "Welt")
        self.assertTrue("Hallo" in layer.userData)

    def test_smartComponentPoleMapping(self):
        # http://docu.glyphsapp.com/#smartComponentPoleMapping
        # Read some data from the file
        shoulder = self.font.glyphs["_part.shoulder"]
        for layer in shoulder.layers:
            if layer.name == "NarrowShoulder":
                mapping = layer.smartComponentPoleMapping
                self.assertIsNotNone(mapping)
                # crotchDepth is at the top pole
                self.assertEqual(2, mapping["crotchDepth"])
                # shoulderWidth is at the bottom pole
                self.assertEqual(1, mapping["shoulderWidth"])

        # Exercise the getter/setter
        layer = self.layer
        self.assertDict(layer.smartComponentPoleMapping)
        self.assertFalse("crotchDepth" in layer.smartComponentPoleMapping)
        layer.smartComponentPoleMapping["crotchDepth"] = 2
        self.assertTrue("crotchDepth" in layer.smartComponentPoleMapping)
        layer.smartComponentPoleMapping = {"shoulderWidth": 1}
        self.assertFalse("crotchDepth" in layer.smartComponentPoleMapping)
        self.assertEqual(1, layer.smartComponentPoleMapping["shoulderWidth"])

    # TODO: Methods
    # copyDecomposedLayer()
    # decomposeComponents()
    # compareString()
    # connectAllOpenPaths()
    # syncMetrics()
    # correctPathDirection()
    # removeOverlap()
    # roundCoordinates()
    # addNodesAtExtremes()
    # applyTransform()
    # beginChanges()
    # endChanges()
    # cutBetweenPoints()
    # intersectionsBetweenPoints()
    # addMissingAnchors()
    # clearSelection()
    # swapForegroundWithBackground()
    # reinterpolate()
    # clear()


class GSComponentFromFileTest(GSObjectsTestCase):
    def setUp(self):
        super().setUp()
        self.glyph = self.font.glyphs["adieresis"]
        self.layer = self.glyph.layers[0]
        self.component = self.layer.components[0]

    def test_repr(self):
        component = self.component
        self.assertIsNotNone(component.__repr__())
        self.assertEqual(repr(component), '<GSComponent "a" x=0.0 y=0.0>')

    def test_delete_and_add(self):
        layer = self.layer
        self.assertEqual(len(layer.components), 2)
        layer.components = []
        self.assertEqual(len(layer.components), 0)
        layer.components.append(GSComponent("a"))
        self.assertIsNotNone(layer.components[0].__repr__())
        self.assertEqual(len(layer.components), 1)
        layer.components.append(GSComponent("dieresis"))
        self.assertEqual(len(layer.components), 2)
        layer.components = [GSComponent("a"), GSComponent("dieresis")]
        self.assertEqual(len(layer.components), 2)
        layer.components = []
        layer.components.extend([GSComponent("a"), GSComponent("dieresis")])
        self.assertEqual(len(layer.components), 2)
        newComponent = GSComponent("dieresis")
        layer.components.insert(0, newComponent)
        self.assertEqual(newComponent, layer.components[0])
        layer.components.remove(layer.components[0])
        self.assertEqual(len(layer.components), 2)

    def test_position(self):
        self.assertIsInstance(self.component.position, Point)

    def test_componentName(self):
        self.assertUnicode(self.component.componentName)

    def test_component(self):
        self.assertIsInstance(self.component.component, GSGlyph)

    def test_rotation(self):
        self.assertFloat(self.component.rotation)

    def test_transform(self):
        self.assertIsInstance(self.component.transform, Transform)
        self.assertEqual(len(self.component.transform.value), 6)

    def test_bounds(self):
        self.assertIsInstance(self.component.bounds, Rect)
        bounds = self.component.bounds
        self.assertEqual(bounds.origin.x, 80)
        self.assertEqual(bounds.origin.y, -10)
        self.assertEqual(bounds.size.width, 289)
        self.assertEqual(bounds.size.height, 490)

    def test_moreBounds(self):
        self.component.scale = 1.1
        bounds = self.component.bounds
        self.assertEqual(bounds.origin.x, 88)
        self.assertEqual(bounds.origin.y, -11)
        self.assertEqual(round(bounds.size.width * 10), round(317.9 * 10))
        self.assertEqual(round(bounds.size.height * 10), round(539 * 10))

    # def test_automaticAlignment(self):
    #     self.assertBool(self.component.automaticAlignment)

    def test_anchor(self):
        self.assertString(self.component.anchor)

    def test_smartComponentValues(self):
        glyph = self.font.glyphs["h"]
        stem, shoulder = glyph.layers[0].components
        self.assertEqual(100, stem.smartComponentValues["height"])
        self.assertEqual(-80.20097, shoulder.smartComponentValues["crotchDepth"])

        self.assertNotIn("shoulderWidth", shoulder.smartComponentValues)

        self.assertNotIn("somethingElse", shoulder.smartComponentValues)

    # bezierPath?
    # componentLayer()


class GSGuideTest(unittest.TestCase):
    def test_repr(self):
        guide = GSGuide()
        self.assertEqual(repr(guide), "<GSGuide x=0.0 y=0.0 angle=0.0>")


class GSAnchorFromFileTest(GSObjectsTestCase):
    def setUp(self):
        super().setUp()
        self.glyph = self.font.glyphs["a"]
        self.layer = self.glyph.layers[0]
        self.anchor = self.layer.anchors[0]

    def test_repr(self):
        anchor = self.anchor
        self.assertEqual(anchor.__repr__(), '<GSAnchor "bottom" x=218.0 y=0.0>')

    def test_name(self):
        anchor = self.anchor
        self.assertUnicode(anchor.name)

    # TODO
    def test_position(self):
        pass


class GSPathFromFileTest(GSObjectsTestCase):
    def setUp(self):
        super().setUp()
        self.glyph = self.font.glyphs["a"]
        self.layer = self.glyph.layers[0]
        self.path = self.layer.paths[0]

    def test_proxy(self):
        layer = self.layer
        path = self.path
        amount = len(layer.paths)
        pathCopy1 = copy.copy(path)
        layer.paths.append(pathCopy1)
        pathCopy2 = copy.copy(pathCopy1)
        layer.paths.extend([pathCopy2])
        self.assertEqual(layer.paths[-2], pathCopy1)
        self.assertEqual(layer.paths[-1], pathCopy2)
        pathCopy3 = copy.copy(pathCopy2)
        layer.paths.insert(0, pathCopy3)
        self.assertEqual(layer.paths[0], pathCopy3)
        layer.paths.remove(layer.paths[0])
        layer.paths.remove(layer.paths[-1])
        layer.paths.remove(layer.paths[-1])
        self.assertEqual(amount, len(layer.paths))

    def test_parent(self):
        path = self.path
        self.assertEqual(path.parent, self.layer)

    def test_nodes(self):
        path = self.path
        self.assertIsNotNone(path.nodes)
        self.assertEqual(len(path.nodes), 44)
        for node in path.nodes:
            self.assertEqual(node.parent, path)
        amount = len(path.nodes)
        newNode = GSNode(Point("{100, 100}"))
        path.nodes.append(newNode)
        self.assertEqual(newNode, path.nodes[-1])
        del path.nodes[-1]
        newNode = GSNode(Point("{20, 20}"))
        path.nodes.insert(0, newNode)
        self.assertEqual(newNode, path.nodes[0])
        path.nodes.remove(path.nodes[0])
        newNode1 = GSNode(Point("{10, 10}"))
        newNode2 = GSNode(Point("{20, 20}"))
        path.nodes.extend([newNode1, newNode2])
        self.assertEqual(newNode1, path.nodes[-2])
        self.assertEqual(newNode2, path.nodes[-1])
        del path.nodes[-2]
        del path.nodes[-1]
        self.assertEqual(amount, len(path.nodes))

    def test_node_position(self):
        n = GSNode()
        n.position = Point("{10, 10}")
        self.assertEqual(n.position.x, 10)
        n.position = (20, 20)
        self.assertEqual(n.position.x, 20)

    # TODO: GSPath.closed

    # bezierPath?

    # TODO:
    # addNodesAtExtremes()
    # applyTransform()
    def test_applyTransform_translate(self):
        pathCopy = copy.copy(self.path)
        pathCopy.applyTransform((1, 0, 0, 1, 50, 25))
        expected = ((402, 172), (402, 93), (364, 32), (262, 32))
        for i, pt in enumerate(expected):
            self.assertEqual(pathCopy.nodes[i].position.x, pt[0])
            self.assertEqual(pathCopy.nodes[i].position.y, pt[1])

    def test_applyTransform_translate_scale(self):
        pathCopy = copy.copy(self.path)
        pathCopy.applyTransform((0.9, 0, 0, 1.2, 50, 25))
        expected = ((367, 201), (367, 107), (333, 33), (241, 33))
        for i, pt in enumerate(expected):
            self.assertAlmostEqual(pathCopy.nodes[i].position.x, pt[0], 0)
            self.assertAlmostEqual(pathCopy.nodes[i].position.y, pt[1], 0)

    def test_applyTransform_skew(self):
        pathCopy = copy.copy(self.path)
        pathCopy.applyTransform((1, 0.1, 0.2, 1, 0, 0))
        expected = ((381, 182), (366, 103), (315, 38), (213, 28))
        for i, pt in enumerate(expected):
            self.assertAlmostEqual(pathCopy.nodes[i].position.x, pt[0], 0)
            self.assertAlmostEqual(pathCopy.nodes[i].position.y, pt[1], 0)

    def test_direction(self):
        self.assertEqual(self.path.direction, -1)

    def test_segments(self):
        oldSegments = self.path.segments
        self.assertEqual(len(self.path.segments), 20)
        self.path.reverse()
        self.assertEqual(len(self.path.segments), 20)
        self.assertEqual(oldSegments[0].nodes[0], self.path.segments[0].nodes[0])

    def test_segments_2(self):
        p = GSPath()
        p.nodes = [
            GSNode((204, 354), "curve"),
            GSNode((198, 353), "offcurve"),
            GSNode((193, 352), "offcurve"),
            GSNode((183, 352), "curve"),
            GSNode((154, 352), "offcurve"),
            GSNode((123, 364), "offcurve"),
            GSNode((123, 384), "curve"),
            GSNode((123, 403), "offcurve"),
            GSNode((148, 419), "offcurve"),
            GSNode((167, 419), "curve"),
            GSNode((190, 419), "offcurve"),
            GSNode((204, 397), "offcurve"),
        ]
        self.assertEqual(len(p.segments), 4)

    def test_segments_3(self):
        p = GSPath()
        p.nodes = [
            GSNode((327, 185), "offcurve"),
            GSNode((299, 210), "offcurve"),
            GSNode((297, 266), "curve"),
            GSNode((294, 351), "offcurve"),
            GSNode((297, 434), "qcurve"),
            GSNode((299, 490), "offcurve"),
            GSNode((328, 515), "offcurve"),
            GSNode((371, 515), "curve"),
            GSNode((414, 515), "offcurve"),
            GSNode((443, 490), "offcurve"),
            GSNode((445, 434), "curve"),
            GSNode((448, 351), "offcurve"),
            GSNode((445, 266), "qcurve"),
            GSNode((443, 210), "offcurve"),
            GSNode((415, 185), "offcurve"),
            GSNode((371, 185), "curve"),
        ]
        self.assertEqual(len(p.segments), 6)

    def test_segments_4(self):
        p = GSPath()
        p.nodes = [
            GSNode((327, 185), "line"),
            GSNode((297, 266), "line"),
            GSNode((371, 185), "line"),
        ]
        self.assertEqual(len(p.segments), 3)
        p.closed = False
        self.assertEqual(len(p.segments), 2)
        self.assertEqual(p.segments[0][0].x, 327)

    def test_segments_5(self):
        p = GSPath()
        p.nodes = [
            GSNode((250, 2000), "offcurve"),
            GSNode((250, 1000), "offcurve"),
            GSNode((250, 900), "curve"),
            GSNode((250, 500), "offcurve"),
            GSNode((250, 50), "offcurve"),
            GSNode((250, 0), "curve"),
            GSNode((250, 1700), "curve"),
            # Yes, this is bad construction but we shouldn't
            # infinite loop
        ]
        self.assertEqual(len(p.segments), 3)

    def test_bounds(self):
        bounds = self.path.bounds
        self.assertEqual(bounds.origin.x, 80)
        self.assertEqual(bounds.origin.y, -10)
        self.assertEqual(bounds.size.width, 289)
        self.assertEqual(bounds.size.height, 490)


class GSNodeFromFileTest(GSObjectsTestCase):
    def setUp(self):
        super().setUp()
        self.glyph = self.font.glyphs["a"]
        self.layer = self.glyph.layers[0]
        self.path = self.layer.paths[0]
        self.node = self.path.nodes[0]

    def test_repr(self):
        self.assertIsNotNone(self.node.__repr__())

    def test_position(self):
        self.assertIsInstance(self.node.position, Point)

    def test_type(self):
        self.assertTrue(self.node.type in [LINE, CURVE, OFFCURVE])

    def test_smooth(self):
        self.assertBool(self.node.smooth)

    def test_index(self):
        self.assertInteger(self.node.index)
        self.assertEqual(self.path.nodes[0].index, 0)
        self.assertEqual(self.path.nodes[-1].index, 43)

    def test_nextNode(self):
        self.assertEqual(type(self.path.nodes[-1].nextNode), GSNode)
        self.assertEqual(self.path.nodes[-1].nextNode, self.path.nodes[0])

    def test_prevNode(self):
        self.assertEqual(type(self.path.nodes[0].prevNode), GSNode)
        self.assertEqual(self.path.nodes[0].prevNode, self.path.nodes[-1])

    def test_name(self):
        self.assertEqual(self.node.name, "Hello")

    def test_userData(self):
        self.assertEqual("1", self.node.userData["rememberToMakeCoffee"])

    def test_makeNodeFirst(self):
        oldAmount = len(self.path.nodes)
        oldSecondNode = self.path.nodes[3]
        self.path.nodes[3].makeNodeFirst()
        self.assertEqual(oldAmount, len(self.path.nodes))
        self.assertEqual(oldSecondNode, self.path.nodes[0])

    def test_toggleConnection(self):
        oldConnection = self.node.smooth
        self.node.toggleConnection()
        self.assertEqual(oldConnection, not self.node.smooth)


class GSCustomParameterTest(unittest.TestCase):
    def test_plistValue_string(self):
        test_string = "Some Value"
        param = GSCustomParameter("New Parameter", test_string)
        self.assertEqual(
            param.plistValue(), '{\nname = "New Parameter";\nvalue = "Some Value";\n}'
        )

    def test_plistValue_list(self):
        test_list = [1, 2.5, {"key1": "value1"}]
        param = GSCustomParameter("New Parameter", test_list)
        self.assertEqual(
            param.plistValue(),
            '{\nname = "New Parameter";\nvalue = (\n1,\n2.5,'
            "\n{\nkey1 = value1;\n}\n);\n}",
        )

    def test_plistValue_dict(self):
        test_dict = {"key1": "value1", "key2": "value2"}
        param = GSCustomParameter("New Parameter", test_dict)
        self.assertEqual(
            param.plistValue(),
            '{\nname = "New Parameter";\nvalue = {\nkey1 = value1;'
            "\nkey2 = value2;\n};\n}",
        )


class GSBackgroundLayerTest(unittest.TestCase):
    """Goal: forbid in glyphsLib all the GSLayer.background APIs that don't
    work in Glyphs.app, so that the code we write for glyphsLib is sure to
    work in Glyphs.app
    """

    def setUp(self):
        self.layer = GSLayer()
        self.bg = self.layer.background

    def test_get_GSLayer_background(self):
        """It should always return a GSLayer (actually a GSBackgroundLayer but
        it's a subclass of GSLayer so it's ok)
        """
        self.assertIsInstance(self.bg, GSLayer)
        bg2 = self.layer.background
        self.assertEqual(self.bg, bg2)

    def test_set_GSLayer_background(self):
        """It should raise because it behaves strangely in Glyphs.app.
        The only way to modify a background layer in glyphsLib is to get it
        from a GSLayer object.
        """
        with pytest.raises(AttributeError):
            self.layer.background = GSLayer()

    def test_get_GSLayer_foreground(self):
        """It should raise AttributeError, as in Glyphs.app"""
        with pytest.raises(AttributeError):
            self.layer.foreground

    def test_set_GSLayer_foreground(self):
        with pytest.raises(AttributeError):
            self.layer.foreground = GSLayer()

    def test_get_GSBackgroundLayer_background(self):
        """It should always return None, as in Glyphs.app"""
        self.assertIsNone(self.bg.background)

    def test_set_GSBackgroundLayer_background(self):
        """It should raise because it should not be possible."""
        with pytest.raises(AttributeError):
            self.bg.background = GSLayer()

    def test_get_GSBackgroundLayer_foreground(self):
        """It should return the foreground layer.

        WARNING: currently in Glyphs.app it is not implemented properly and it
        returns some Objective C function.
        """
        self.assertEqual(self.layer, self.bg.foreground)

    def test_set_GSBackgroundLayer_foreground(self):
        """It should raise AttributeError, because it would be too complex to
        implement properly and anyway in Glyphs.app it returns some Objective C
        function.
        """
        with pytest.raises(AttributeError):
            self.bg.foreground = GSLayer()


class segmentTest(unittest.TestCase):
    def test_bbox_bug(self):
        seg = segment(
            [Point(529, 223), Point(447, 456), Point(285, 177), Point(521, 367)]
        )
        bbox = seg.bbox()
        self.assertAlmostEqual(bbox[0], 398.1222655016518)
        self.assertAlmostEqual(bbox[1], 223)
        self.assertAlmostEqual(bbox[2], 529)
        self.assertAlmostEqual(bbox[3], 367)


class FontGlyphsProxyTest(unittest.TestCase):
    def setUp(self):
        self.font = GSFont(TESTFILE_PATH)

    def test_remove_glyphs(self):
        assert self.font.glyphs[0].name == "A"
        del self.font.glyphs[0]
        assert self.font.glyphs[0].name != "A"

        assert self.font.glyphs["Adieresis"].name == "Adieresis"
        del self.font.glyphs["Adieresis"]
        assert self.font.glyphs["Adieresis"] is None

        with pytest.raises(KeyError):
            del self.font.glyphs["xxxzzz"]

        with pytest.raises(KeyError):
            del self.font.glyphs[self.font]

    def test_append_after_setter_keeps_index_consistent(self):
        """Appending a glyph after assigning font.glyphs must not leave
        the name index in an incomplete state where only the appended
        glyph is findable by name."""
        font = GSFont()
        # Assign glyphs via the property setter
        g1 = GSGlyph("A")
        g2 = GSGlyph("B")
        font.glyphs = [g1, g2]

        # Now append a third glyph
        g3 = GSGlyph("C")
        font.glyphs.append(g3)

        # All three must be findable by name
        assert font.glyphs["A"] is not None, "glyph 'A' not found after setter+append"
        assert font.glyphs["B"] is not None, "glyph 'B' not found after setter+append"
        assert font.glyphs["C"] is not None, "glyph 'C' not found after setter+append"

    def test_setitem_invalidates_index(self):
        """Replacing a glyph by index must invalidate the name index so
        the old name is no longer found and the new name is."""
        font = GSFont()
        font.glyphs = [GSGlyph("A"), GSGlyph("B")]

        # Force the index to be built
        assert font.glyphs["A"] is not None

        # Replace glyph at index 0
        font.glyphs[0] = GSGlyph("Z")

        assert font.glyphs["Z"] is not None, "glyph 'Z' not found after replacement"
        assert font.glyphs["A"] is None, "glyph 'A' still found after replacement"


class FontClassesProxyTest(unittest.TestCase):
    def setUp(self):
        self.font = GSFont(TESTFILE_PATH)

    def test_indxing_by_name(self):
        assert "Languagesystems" in self.font.featurePrefixes
        assert "c2sc_source" in self.font.classes
        assert "aalt" in self.font.features

        assert "XXXX" not in self.font.featurePrefixes
        assert "XXXX" not in self.font.classes
        assert "XXXX" not in self.font.features

        assert self.font.featurePrefixes["Languagesystems"] in self.font.featurePrefixes
        assert self.font.classes["c2sc_source"] in self.font.classes
        assert self.font.features["aalt"] in self.font.features


if __name__ == "__main__":
    unittest.main()



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging

from fontTools.pens.basePen import MissingComponentError
from fontTools.pens.recordingPen import DecomposingRecordingPen
from glyphsLib.classes import GSBackgroundLayer
from glyphsLib.types import Transform

from .smart_components import instantiate_smart_component
from .constants import GLYPHS_PREFIX, COMPONENT_INFO_KEY, SMART_COMPONENT_AXES_LIB_KEY

logger = logging.getLogger(__name__)


def to_ufo_components(self, ufo_glyph, layer):
    """Draw .glyphs components onto a pen, adding them to the parent glyph."""

    # NOTE: The UFO v3 and Glyphs data model have incompatible component reference
    # semantics. UFO components always point to a glyph in the same layer, Glyphs
    # components in a ...:
    #  - master layer: point to glyphs in the same master layer.
    #  - non-master layer: point to glyphs in the layer with the same name and fall
    #    back to glyphs in the associated master layer
    # There are some valid use-cases for components in non-master layers, and doing it
    # thoroughly correctly is time-consuming, so we're decomposing just the background
    # layer components as a band-aid.
    if layer.components and isinstance(layer, GSBackgroundLayer):
        logger.warning(
            f"Glyph '{ufo_glyph.name}': All components of the background layer of "
            f"'{layer.foreground.name}' will be decomposed."
        )
        to_ufo_components_nonmaster_decompose(self, ufo_glyph, layer)
        return

    pen = ufo_glyph.getPointPen()
    for index, component in enumerate(layer.components):
        component_name = component.name
        if layer._is_color_palette_layer():
            # Glyphs handles components for color layers in a special way. If
            # the component glyph has color layers of its own, the component
            # use the first color layer with the same color index, otherwise it
            # fallback to the default layer. We try to do that here as well.
            font = layer.parent.parent
            component_glyph = font.glyphs[component_name]
            color_layers = [
                l
                for l in component_glyph.layers
                if l._is_color_palette_layer()
                and l.associatedMasterId == layer.associatedMasterId
            ]
            for i, l in enumerate(color_layers):
                if l._color_palette_index() == layer._color_palette_index():
                    if l.layerId != l.associatedMasterId:
                        # If it is not a master layer, we rename it in
                        # _to_ufo_color_palette_layers(), so we reference the
                        # same name here.
                        component_name += f".color{i}"
                    break
        # XXX We may also want to test here if we're compiling a font (and decompose
        # if so) or changing the representation format (in which case we leave it
        # as a component and save the smart component values).
        # See https://github.com/googlefonts/glyphsLib/pull/822
        if component.component and component.component.smartComponentAxes:
            instantiate_smart_component(self, layer, component, pen)
        else:
            pen.addComponent(component_name, component.transform)

        if not (component.anchor or component.alignment):
            continue

        component_info = {"name": component.name, "index": index}
        if component.anchor:
            component_info["anchor"] = component.anchor
        if component.alignment:
            component_info["alignment"] = component.alignment

        if COMPONENT_INFO_KEY not in ufo_glyph.lib:
            ufo_glyph.lib[COMPONENT_INFO_KEY] = []
        ufo_glyph.lib[COMPONENT_INFO_KEY].append(component_info)

    # data related to components that is not stored in ComponentInfo is
    # stored in lists of booleans. each list's elements correspond to the
    # components in order.
    for key in ["locked", "smartComponentValues"]:
        values = [getattr(c, key) for c in layer.components]
        if any(values):
            ufo_glyph.lib[_lib_key(key)] = values


def to_ufo_components_nonmaster_decompose(self, ufo_glyph, layer):
    """Draw decomposed .glyphs background and non-master layers with a pen,
    adding them to the parent glyph."""

    if isinstance(layer, GSBackgroundLayer):
        layer_id = layer.foreground.layerId
        layer_master_id = layer.foreground.associatedMasterId
    else:
        layer_id = layer.layerId
        layer_master_id = layer.associatedMasterId

    if layer_id in self._glyph_sets:
        layers = self._glyph_sets[layer_id]
    else:
        if layer_id == layer_master_id:
            # Is a master layer.
            layers = self._glyph_sets[layer_id] = {
                g.name: l
                for g in layer.parent.parent.glyphs
                for l in g.layers
                if l.layerId == layer_id
            }
        else:
            # Is a non-master layer.
            layers_nonmaster = {
                g.name: l
                for g in layer.parent.parent.glyphs
                for l in g.layers
                if l.name == layer.name
            }
            layers_master = {
                g.name: l
                for g in layer.parent.parent.glyphs
                for l in g.layers
                if l.layerId == layer_master_id
            }
            layers = self._glyph_sets[layer_id] = {
                **layers_master,
                **layers_nonmaster,
            }

    rpen = DecomposingRecordingPen(glyphSet=layers)
    for component in layer.components:
        try:
            component.draw(rpen)
        except MissingComponentError as e:
            raise MissingComponentError(
                f"Glyph '{ufo_glyph.name}', background layer: component "
                f"'{component.name}' points to a non-existent glyph."
            ) from e
    rpen.replay(ufo_glyph.getPen())


def to_glyphs_components(self, ufo_glyph, layer):
    for comp in ufo_glyph.components:
        component = self.glyphs_module.GSComponent(comp.baseGlyph)
        component.transform = Transform(*comp.transformation)
        layer.components.append(component)

    for key in ["alignment", "locked", "smartComponentValues"]:
        if _lib_key(key) not in ufo_glyph.lib:
            continue
        # FIXME: (jany) move to using component identifiers for robustness
        # "alignment" is read, but not written for source backwards compatibility.
        values = ufo_glyph.lib[_lib_key(key)]
        for component, value in zip(layer.components, values):
            if value is not None:
                setattr(component, key, value)

    if COMPONENT_INFO_KEY in ufo_glyph.lib:
        for index, component_info in enumerate(ufo_glyph.lib[COMPONENT_INFO_KEY]):
            if "index" not in component_info or "name" not in component_info:
                logger.warning(
                    "Glyph %s, layer %s: The ComponentInfo at index %s is missing "
                    "index and/or name keys. Skipping, component properties will be "
                    "lost.",
                    ufo_glyph.name,
                    layer.name,
                    index,
                )
                continue

            component_index = component_info["index"]
            try:
                component = layer.components[component_index]
            except IndexError:
                logger.warning(
                    "Glyph %s, layer %s: The ComponentInfo at index %s is referencing "
                    "a component that does not exist. Skipping, component properties "
                    "will be lost.",
                    ufo_glyph.name,
                    layer.name,
                    index,
                )
                continue

            if component.name == component_info["name"]:
                if "anchor" in component_info:
                    component.anchor = component_info["anchor"]
                if "alignment" in component_info:
                    component.alignment = component_info["alignment"]
            else:
                logger.warning(
                    "Glyph %s, layer %s: The ComponentInfo at index %s says the "
                    "component at index %s is named '%s', but it is actually named "
                    "'%s'. Skipping, component properties will be lost.",
                    ufo_glyph.name,
                    layer.name,
                    index,
                    component_index,
                    component_info["name"],
                    component.name,
                )


def _lib_key(key):
    key = key[0].upper() + key[1:]
    return f"{GLYPHS_PREFIX}components{key}"


AXES_LIB_KEY = GLYPHS_PREFIX + "smartComponentAxes"


def to_ufo_smart_component_axes(self, ufo_glyph, glyph):
    def _to_ufo_axis(axis):
        return {
            "name": axis.name,
            "bottomName": axis.bottomName,
            "bottomValue": axis.bottomValue,
            "topName": axis.topName,
            "topValue": axis.topValue,
        }

    if glyph.smartComponentAxes:
        ufo_glyph.lib[SMART_COMPONENT_AXES_LIB_KEY] = [
            _to_ufo_axis(a) for a in glyph.smartComponentAxes
        ]


def to_glyphs_smart_component_axes(self, ufo_glyph, glyph):
    def _to_glyphs_axis(axis):
        res = self.glyphs_module.GSSmartComponentAxis()
        res.name = axis["name"]
        res.bottomName = axis["bottomName"]
        res.bottomValue = axis["bottomValue"]
        res.topValue = axis["topValue"]
        res.topName = axis["topName"]
        return res

    if SMART_COMPONENT_AXES_LIB_KEY in ufo_glyph.lib:
        glyph.smartComponentAxes = [
            _to_glyphs_axis(a) for a in ufo_glyph.lib[SMART_COMPONENT_AXES_LIB_KEY]
        ]



import py
import pytest


@pytest.fixture
def datadir(request):
    return py.path.local(py.path.local(__file__).dirname).join("data")


@pytest.fixture(scope="session", params=["defcon", "ufoLib2"])
def ufo_module(request):
    return pytest.importorskip(request.param)


# Provide a --run-regression-tests CLI option to run slow regression tests separately.
def pytest_addoption(parser):
    parser.addoption(
        "--run-regression-tests",
        action="store_true",
        help="Run (slow) regression tests",
    )


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "regression_test: mark test as a (slow) regression test"
    )


def pytest_collection_modifyitems(config, items):
    if config.getoption("--run-regression-tests"):
        # --run-regression-tests given in cli: do not skip slow tests
        return
    skip_regression_test = pytest.mark.skip(
        reason="need --run-regression-tests option to run"
    )
    for item in items:
        if "regression_test" in item.keywords:
            item.add_marker(skip_regression_test)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import re

from glyphsLib.util import cast_to_number_or_bool, reverse_cast_to_number_or_bool

logger = logging.getLogger(__name__)


def parse_glyphs_filter(filter_str, is_pre=False):
    """Parses glyphs custom filter string into a dict object that
    ufo2ft can consume.

     Reference:
         ufo2ft: https://github.com/googlefonts/ufo2ft
         Glyphs 2.3 Handbook July 2016, p184

     Args:
         filter_str - a string of glyphs app filter

     Return:
         A dictionary contains the structured filter.
         Return None if parse failed.
    """
    elements = filter_str.split(";")

    if elements[0] == "":
        logger.error(
            "Failed to parse glyphs filter, expecting a filter name: \
             %s",
            filter_str,
        )
        return None

    result = {"name": elements[0]}
    for idx, elem in enumerate(elements[1:]):
        if not elem:
            # skip empty arguments
            continue
        if ":" in elem:
            # Key value pair
            key, value = elem.split(":", 1)
            if key.lower() in ["include", "exclude"]:
                if idx != len(elements[1:]) - 1:
                    logger.error(
                        "{} can only present as the last argument in the filter. "
                        "{} is ignored.".format(key, elem)
                    )
                    continue
                result[key.lower()] = re.split("[ ,]+", value)
            else:
                if "kwargs" not in result:
                    result["kwargs"] = {}
                result["kwargs"][key] = cast_to_number_or_bool(value)
        else:
            if "args" not in result:
                result["args"] = []
            result["args"].append(cast_to_number_or_bool(elem))
    if is_pre:
        result["pre"] = True
    return result


def write_glyphs_filter(result):
    elements = [result["name"]]
    if "args" in result:
        for arg in result["args"]:
            elements.append(reverse_cast_to_number_or_bool(arg))
    if "kwargs" in result:
        for key, arg in result["kwargs"].items():
            if key.lower() not in ("include", "exclude"):
                elements.append(key + ":" + reverse_cast_to_number_or_bool(arg))
        for key, arg in result["kwargs"].items():
            if key.lower() in ("include", "exclude"):
                elements.append(key + ":" + reverse_cast_to_number_or_bool(arg))
    for key, arg in result.items():
        if key.lower() in ("include", "exclude"):
            elements.append(key + ":" + ",".join(arg))
    return ";".join(elements)



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import copy
import logging

from glyphsLib import classes, glyphdata

from .builders import UFOBuilder, GlyphsBuilder
from .transformations import TRANSFORMATIONS, TRANSFORMATION_CUSTOM_PARAMS

logger = logging.getLogger(__name__)


def to_ufos(
    font,
    include_instances=False,
    family_name=None,
    propagate_anchors=None,
    ufo_module=None,
    minimize_glyphs_diffs=False,
    generate_GDEF=True,
    store_editor_state=True,
    write_skipexportglyphs=False,
    expand_includes=False,
    minimal=False,
    glyph_data=None,
    preserve_original=False,
):
    """Take a GSFont object and convert it into one UFO per master.

    Takes in data as Glyphs.app-compatible classes, as documented at
    https://docu.glyphsapp.com/. The input ``GSFont`` object is modified
    unless ``preserve_original`` is true.

    If include_instances is True, also returns the parsed instance data.

    If family_name is provided, the master UFOs will be given this name and
    only instances with this name will be returned.

    If generate_GDEF is True, write a `table GDEF {...}` statement in the
    UFO's features.fea, containing GlyphClassDef and LigatureCaretByPos.

    If expand_includes is True, resolve include statements in the GSFont features
    and inline them in the UFO features.fea.

    If minimal is True, it is assumed that the UFOs will only be used in
    font production, and unnecessary steps (e.g. converting background layers)
    will be skipped.

    If preserve_original is True, this works on a copy of the font object
    to avoid modifying the original object.

    The optional glyph_data parameter takes a list of GlyphData.xml paths or
    a pre-parsed GlyphData object that overrides the default one.
    """
    if preserve_original:
        font = copy.deepcopy(font)
    if glyph_data is not None and not isinstance(glyph_data, glyphdata.GlyphData):
        glyph_data = glyphdata.GlyphData.from_files(*glyph_data)
    font = preflight_glyphs(
        font, glyph_data=glyph_data, do_propagate_all_anchors=propagate_anchors
    )
    builder = UFOBuilder(
        font,
        ufo_module=ufo_module,
        family_name=family_name,
        minimize_glyphs_diffs=minimize_glyphs_diffs,
        generate_GDEF=generate_GDEF,
        store_editor_state=store_editor_state,
        write_skipexportglyphs=write_skipexportglyphs,
        expand_includes=expand_includes,
        minimal=minimal,
        glyph_data=glyph_data,
    )

    result = list(builder.masters)

    if include_instances:
        return result, builder.instance_data
    return result


def to_designspace(
    font,
    family_name=None,
    instance_dir=None,
    propagate_anchors=None,
    ufo_module=None,
    minimize_glyphs_diffs=False,
    generate_GDEF=True,
    store_editor_state=True,
    write_skipexportglyphs=False,
    expand_includes=False,
    minimal=False,
    glyph_data=None,
    preserve_original=False,
):
    """Take a GSFont object and convert it into a Designspace Document + UFOS.
    The UFOs are available as the attribute `font` of each SourceDescriptor of
    the DesignspaceDocument:

        ufos = [source.font for source in designspace.sources]

    The input object is modified unless ``preserve_original`` is true.

    The designspace and the UFOs are not written anywhere by default, they
    are all in-memory. If you want to write them to the disk, consider using
    the `filename` attribute of the DesignspaceDocument and of its
    SourceDescriptor as possible file names.

    Takes in data as Glyphs.app-compatible classes, as documented at
    https://docu.glyphsapp.com/

    If include_instances is True, also returns the parsed instance data.

    If family_name is provided, the master UFOs will be given this name and
    only instances with this name will be returned.

    If generate_GDEF is True, write a `table GDEF {...}` statement in the
    UFO's features.fea, containing GlyphClassDef and LigatureCaretByPos.

    If preserve_original is True, this works on a copy of the font object
    to avoid modifying the original object.

    The optional glyph_data parameter takes a list of GlyphData.xml paths or
    a pre-parsed GlyphData object that overrides the default one.
    """
    if preserve_original:
        font = copy.deepcopy(font)
    if glyph_data is not None and not isinstance(glyph_data, glyphdata.GlyphData):
        glyph_data = glyphdata.GlyphData.from_files(*glyph_data)
    font = preflight_glyphs(
        font, glyph_data=glyph_data, do_propagate_all_anchors=propagate_anchors
    )
    builder = UFOBuilder(
        font,
        ufo_module=ufo_module,
        family_name=family_name,
        instance_dir=instance_dir,
        use_designspace=True,
        minimize_glyphs_diffs=minimize_glyphs_diffs,
        generate_GDEF=generate_GDEF,
        store_editor_state=store_editor_state,
        write_skipexportglyphs=write_skipexportglyphs,
        expand_includes=expand_includes,
        minimal=minimal,
        glyph_data=glyph_data,
    )
    return builder.designspace


def preflight_glyphs(font, *, glyph_data=None, **flags):
    """Run a set of transformations over a GSFont object to make
    it easier to convert to UFO; resolve all the "smart stuff".

    Currently, the transformations are:
        - `propagate_all_anchors`: copy anchors from components to their parent

    More transformations may be added in the future.

    Some transformations may have custom parameters that can be set in the
    font. For example, the `propagate_all_anchors` transformation can be
    disabled by setting the custom parameter "Propagate Anchors" to False
    (see `TRANSFORMATION_CUSTOM_PARAMS`).

    Args:
        font: a GSFont object
        glyph_data: an optional GlyphData object associating various properties to
            glyph names (e.g. category) that overrides the default one
        **flags: a set of boolean flags to enable/disable specific transformations,
            named `do_<transformation_name>`, e.g. `do_propagate_all_anchors=False`
            will disable the propagation of anchors.

    Returns:
        the modified GSFont object
    """

    for transform in TRANSFORMATIONS:
        do_transform = flags.pop("do_" + transform.__name__, None)
        if do_transform is True:
            pass
        elif do_transform is False:
            continue
        elif do_transform is None:
            if transform in TRANSFORMATION_CUSTOM_PARAMS:
                param = TRANSFORMATION_CUSTOM_PARAMS[transform]
                if not font.customParameters.get(param.name, param.default):
                    continue
        else:
            raise ValueError(f"Invalid value for do_{transform.__name__}")
        logger.info(f"Running '{transform.__name__}' transformation")
        transform(font, glyph_data=glyph_data)
    if flags:
        logger.warning(f"preflight_glyphs has unused `flags` arguments: {flags}")
    return font


def to_glyphs(
    ufos_or_designspace,
    glyphs_module=classes,
    ufo_module=None,
    minimize_ufo_diffs=False,
    expand_includes=False,
):
    """
    Take a list of UFOs or a single DesignspaceDocument with attached UFOs
    and converts it into a GSFont object.

    The GSFont object is in-memory, it's up to the user to write it to the disk
    if needed.

    This should be the inverse function of `to_ufos` and `to_designspace`,
    so we should have to_glyphs(to_ufos(font)) == font
    and also to_glyphs(to_designspace(font)) == font
    """
    if hasattr(ufos_or_designspace, "sources"):
        builder = GlyphsBuilder(
            designspace=ufos_or_designspace,
            glyphs_module=glyphs_module,
            ufo_module=ufo_module,
            minimize_ufo_diffs=minimize_ufo_diffs,
            expand_includes=expand_includes,
        )
    else:
        builder = GlyphsBuilder(
            ufos=ufos_or_designspace,
            glyphs_module=glyphs_module,
            ufo_module=ufo_module,
            minimize_ufo_diffs=minimize_ufo_diffs,
            expand_includes=expand_includes,
        )
    return builder.font



# Copyright 2016 Georg Seifert. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http: #www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import glyphsLib.classes
from glyphsLib.types import floatToString5
import logging
import datetime
from collections import OrderedDict
from io import StringIO

"""
    Usage

    >> fp = open('Path/to/File.glyphs', 'w')
    >> writer = Writer(fp)
    >> writer.write(font)
    >> fp.close()
"""

logger = logging.getLogger(__name__)


class Writer:
    def __init__(self, fp, format_version=2):
        # figure out whether file object expects bytes or unicodes
        try:
            fp.write(b"")
        except TypeError:
            fp.write("")  # this better not fail...
            # file already accepts unicodes; use it directly
            self.file = fp
        else:
            # file expects bytes; wrap it in a UTF-8 codecs.StreamWriter
            import codecs

            self.file = codecs.getwriter("utf-8")(fp)
        self.format_version = format_version

    def write(self, rootObject):
        self.writeDict(rootObject)
        self.file.write("\n")

    def writeDict(self, dictValue):
        if hasattr(dictValue, "_serialize_to_plist"):
            self.file.write("{\n")
            dictValue._serialize_to_plist(self)
            self.file.write("}")
            return
        self.file.write("{\n")
        keys = dictValue.keys()
        if not isinstance(dictValue, OrderedDict):
            keys = sorted(keys)
        for key in keys:
            try:
                if isinstance(dictValue, (dict, OrderedDict)):
                    value = dictValue[key]
                else:
                    getKey = key
                    value = getattr(dictValue, getKey)
            except AttributeError:
                continue
            if value is None:
                continue
            self.writeKeyValue(key, value)
        self.file.write("}")

    def writeArray(self, arrayValue):
        self.file.write("(\n")
        idx = 0
        length = len(arrayValue)
        if hasattr(arrayValue, "plistArray"):
            arrayValue = arrayValue.plistArray()
        for value in arrayValue:
            self.writeValue(value)
            if idx < length - 1:
                self.file.write(",\n")
            else:
                self.file.write("\n")
            idx += 1
        self.file.write(")")

    def writeUserData(self, userDataValue):
        self.file.write("{\n")
        keys = sorted(userDataValue.keys())
        for key in keys:
            value = userDataValue[key]
            self.writeKey(key)
            self.writeValue(value, key)
            self.file.write(";\n")
        self.file.write("}")

    def writeKeyValue(self, key, value):
        self.writeKey(key)
        self.writeValue(value, key)
        self.file.write(";\n")

    def writeObjectKeyValue(self, d, key, condition=None, keyName=None, default=None):
        value = getattr(d, key)
        if condition == "if_true":
            condition = bool(value)
        if condition is None:
            if default is not None:
                condition = value != default
            else:
                condition = value is not None
        if condition:
            self.writeKey(keyName or key)
            self.writeValue(value, key)
            self.file.write(";\n")

    def writeValue(self, value, forKey=None):
        if hasattr(value, "plistValue"):
            value = value.plistValue(format_version=self.format_version)
            if value is not None:
                self.file.write(value)
        elif forKey in ["color", "strokeColor"] and hasattr(value, "__iter__"):
            # We have to write color tuples on one line or Glyphs 2.4.x
            # misreads it.
            if self.format_version == 2:
                self.file.write(str(tuple(value)))
            else:
                self.file.write("(")
                for ix, v in enumerate(value):
                    self.file.write(str(v))
                    if ix < len(value) - 1:
                        self.file.write(",")
                self.file.write(")")
        elif isinstance(value, (list, glyphsLib.classes.Proxy)):
            if isinstance(value, glyphsLib.classes.UserDataProxy):
                self.writeUserData(value)
            else:
                self.writeArray(value)
        elif isinstance(value, (dict, OrderedDict, glyphsLib.classes.GSBase)):
            self.writeDict(value)
        elif type(value) == float:
            self.file.write(floatToString5(value))
        elif type(value) == int:
            self.file.write(str(value))
        elif type(value) == bytes:
            self.file.write("<" + value.hex() + ">")
        elif type(value) == bool:
            if value:
                self.file.write("1")
            else:
                self.file.write("0")
        elif type(value) == datetime.datetime:
            self.file.write('"%s +0000"' % str(value))
        else:
            value = self.escape_string(str(value), forKey)
            self.file.write(value)

    def writeKey(self, key):
        key = self.escape_string(key, None)
        self.file.write("%s = " % key)

    def escape_string(self, string, forKey):
        if _needs_quotes(string):
            if self.format_version < 3 and forKey != "unicode":
                string = string.replace("\\", "\\\\")
                string = string.replace('"', '\\"')
                string = string.replace("\n", "\\012")
            else:
                string = string.replace("\\", "\\\\")
                string = string.replace('"', '\\"')
            string = '"%s"' % string
        return string


def dump(obj, fp):
    """Write a GSFont object to a .glyphs file.
    'fp' should be a (writable) file object.
    """
    writer = Writer(fp)
    logger.info("Writing .glyphs file")
    if hasattr(obj, "format_version"):
        writer.format_version = obj.format_version
    writer.write(obj)


def dumps(obj):
    """Serialize a GSFont object to a .glyphs file format.
    Return a (unicode) str object.
    """
    fp = StringIO()
    dump(obj, fp)
    return fp.getvalue()


NSPropertyListNameSet = (
    # 0
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    # 16
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    # 32
    False,
    False,
    False,
    False,
    True,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    False,
    True,
    False,
    # 48
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    False,
    False,
    False,
    False,
    False,
    False,
    # 64
    False,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    # 80
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    False,
    False,
    False,
    False,
    True,
    # 96
    False,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    # 112
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    True,
    False,
    False,
    False,
    False,
    False,
)


def _needs_quotes(string):
    if len(string) == 0:
        return True

    # Does it need quotes because of special characters?
    for c in string:
        d = ord(c)
        if d >= 128 or not NSPropertyListNameSet[d]:
            return True

    # Does it need quotes because it could be confused with a number?
    try:
        int(string)
    except ValueError:
        pass
    else:
        return True

    try:
        float(string)
    except ValueError:
        return False
    else:
        return True



#
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from fontTools import designspaceLib

from glyphsLib import to_glyphs, to_designspace


def test_default_master_roundtrips(ufo_module):
    """This test comes from a common scenario while using glyphsLib to go
    back and forth several times with "minimize diffs" in both directions.
    In the end we get UFOs that have information as below, and there was
    a bug that turned "Regular" into "Normal" and changed the default axis
    value.
    """
    thin = ufo_module.Font()
    thin.info.familyName = "CustomFont"
    thin.info.styleName = "Thin"
    thin.lib["com.schriftgestaltung.customParameter.GSFont.Axes"] = [
        {"Name": "Weight", "Tag": "wght"}
    ]
    regular = ufo_module.Font()
    regular.info.familyName = "CustomFont"
    regular.info.styleName = "Regular"
    regular.lib["com.schriftgestaltung.customParameter.GSFont.Axes"] = [
        {"Name": "Weight", "Tag": "wght"}
    ]

    ds = designspaceLib.DesignSpaceDocument()
    weight = ds.newAxisDescriptor()
    weight.tag = "wght"
    weight.name = "Weight"
    weight.minimum = 300
    weight.maximum = 700
    weight.default = 400
    weight.map = [(300, 58), (400, 85), (700, 145)]
    ds.addAxis(weight)

    thinSource = ds.newSourceDescriptor()
    thinSource.font = thin
    thinSource.location = {"Weight": 58}
    thinSource.familyName = "CustomFont"
    thinSource.styleName = "Thin"
    ds.addSource(thinSource)
    regularSource = ds.newSourceDescriptor()
    regularSource.font = regular
    regularSource.location = {"Weight": 85}
    regularSource.familyName = "CustomFont"
    regularSource.styleName = "Regular"
    regularSource.copyFeatures = True
    regularSource.copyGroups = True
    regularSource.copyInfo = True
    regularSource.copyLib = True
    ds.addSource(regularSource)

    font = to_glyphs(ds, minimize_ufo_diffs=True)
    doc = to_designspace(font, minimize_glyphs_diffs=True, ufo_module=ufo_module)

    reg = doc.sources[1]
    assert reg.styleName == "Regular"
    assert reg.font.info.styleName == "Regular"
    assert reg.copyFeatures is True
    assert reg.copyGroups is True
    assert reg.copyInfo is True
    assert reg.copyLib is True



from collections import defaultdict
from functools import partial
from typing import Any

from fontTools import designspaceLib
from fontTools.varLib import FEAVAR_FEATURETAG_LIB_KEY
from fontTools.varLib.featureVars import overlayFeatureVariations

from glyphsLib import util
from .constants import (
    BRACKET_GLYPH_TEMPLATE,
    GLYPHLIB_PREFIX,
)


def to_designspace_bracket_layers(self):
    """Extract bracket layers in a GSGlyph into free-standing UFO glyphs with
    Designspace substitution rules.
    """
    if not self._designspace.axes:
        raise ValueError(
            "Cannot apply bracket layers unless at least one axis is defined."
        )

    # At this stage we will naively emit a designspace rule for every layer.
    bracket_layer_map = defaultdict(partial(defaultdict, list))
    rules = []
    tag_to_name = {axis.tag: axis.name for axis in self._designspace.axes}
    any_non_export_bracket_glyphs = False
    for layer in self.bracket_layers:
        box_with_tag = layer._bracket_info(self._designspace.axes)
        glyph_name = layer.parent.name
        # It'd be nice to use tag for both glyph names and designspace
        # rules, but given designspace refers to axes by name, our hand
        # is forced. (We don't use axis names in glyph names because they
        # might have spaces in them.)
        bracket_layer_map[glyph_name][util.freezedict(box_with_tag)].append(layer)

        # Note this has the side-effect of allocating a .varAltNN glyph name
        bracket_glyph_name = _bracket_glyph_name(self, glyph_name, box_with_tag)

        # for glyphs marked as non-export, we skip adding the substitution rules
        # but we do generate the bracket glyphs in the UFO and append them to
        # the 'public.skipExportGlyphs' list as they might be used as components
        if not layer.parent.export:
            any_non_export_bracket_glyphs = True
            continue

        box_with_name = {tag_to_name[k]: v for k, v in box_with_tag.items()}
        rules.append(([box_with_name], {glyph_name: bracket_glyph_name}))

    # overlayFeatureVariations does exactly what we need in terms of
    # splitting rules into non-overlapping boxes and consolidating
    # multiple substitutions into a single rule.
    for box, mappings in overlayFeatureVariations(rules):
        # Reduce the mappings to a single mapping
        mapping = {}
        for this_mapping in mappings:
            mapping.update(this_mapping)

        self._designspace.addRule(_make_designspace_rule(box, mapping))

    # Set feature for rules
    feat = self.font.customParameters["Feature for Feature Variations"]
    if feat == "rclt":
        self._designspace.rulesProcessingLast = True
    elif feat and feat != "rvrn":
        self._designspace.lib[FEAVAR_FEATURETAG_LIB_KEY] = feat

    # Finally, copy bracket layers to their own glyphs.
    copy_bracket_layers_to_ufo_glyphs(self, bracket_layer_map)

    # we need to update the skipExportGlyphs list if there were any bracket glyphs
    # marked as non-export. We do it in-place because a reference to the same list
    # object is shared with the master UFO's 'public.skipExportGlyphs' and we
    # want to keep them in sync.
    if self.write_skipexportglyphs and any_non_export_bracket_glyphs:
        self._designspace.lib["public.skipExportGlyphs"][:] = sorted(
            self.skip_export_glyphs
        )

    # re-generate the GDEF table since we have added new BRACKET glyphs, which may
    # also need to be included: https://github.com/googlefonts/glyphsLib/issues/578
    if self.generate_GDEF:
        self.regenerate_gdef()


def copy_bracket_layers_to_ufo_glyphs(self, bracket_layer_map):
    font = self.font
    master_ids = {m.id for m in font.masters}
    # when a glyph master layer doesn't have an explicitly associated bracket layer
    # for any crosspoint locations, we assume the master layer itself will be
    # used implicitly as bracket layer for that location. See "Switching Only One
    # Master" paragraph in "Alternating Glyph Shapes" tutorial at:
    # https://glyphsapp.com/tutorials/alternating-glyph-shapes
    implicit_bracket_layers = set()
    # collect all bracket glyph names for resolving composite references
    bracket_glyphs = set()

    for glyph_name, glyph_bracket_layers in bracket_layer_map.items():
        glyph = font.glyphs[glyph_name]
        for frozenbox, bracket_layers in glyph_bracket_layers.items():
            box = dict(frozenbox)
            for missing_master_layer_id in master_ids.difference(
                bl.associatedMasterId for bl in bracket_layers
            ):
                master_layer = glyph.layers[missing_master_layer_id]
                bracket_layers.append(master_layer)
                implicit_bracket_layers.add(id(master_layer))

            bracket_glyphs.add(_bracket_glyph_name(self, glyph_name, box))

    for glyph_name, glyph_bracket_layers in bracket_layer_map.items():
        for frozenbox, layers in glyph_bracket_layers.items():
            box = dict(frozenbox)
            ufo_glyph_name = _bracket_glyph_name(self, glyph_name, box)
            for layer in layers:
                layer_id = layer.associatedMasterId or layer.layerId
                ufo_font = self._sources[layer_id].font
                ufo_layer = ufo_font.layers.defaultLayer
                ufo_glyph = ufo_layer.newGlyph(ufo_glyph_name)
                self.to_ufo_glyph(ufo_glyph, layer, layer.parent)
                ufo_glyph.unicodes = []  # Avoid cmap interference
                # implicit bracket layers have no distinct name, they are simply
                # references to master layers; the empty string is a signal when
                # roundtripping back to Glyphs to skip the duplicate layers.
                ufo_glyph.lib[GLYPHLIB_PREFIX + "_originalLayerName"] = (
                    "" if id(layer) in implicit_bracket_layers else layer.name
                )
                # swap components if base glyph contains matching bracket layers.
                for comp in ufo_glyph.components:
                    bracket_comp_name = _bracket_glyph_name(self, comp.baseGlyph, box)
                    if bracket_comp_name in bracket_glyphs:
                        comp.baseGlyph = bracket_comp_name
                # Update kerning groups and pairs, bracket glyphs inherit the
                # parent's kerning.
                _expand_kerning_to_brackets(glyph_name, ufo_glyph_name, ufo_font)


def _bracket_glyph_name(self, glyph_name, box):
    if box not in self.alternate_names_map[glyph_name]:
        self.alternate_names_map[glyph_name].append(box)
    description = "varAlt%02i" % (1 + self.alternate_names_map[glyph_name].index(box))
    return BRACKET_GLYPH_TEMPLATE.format(glyph_name=glyph_name, description=description)


def _make_designspace_rule(box, mapping):
    description = ".".join(
        f"{name}_{min}_{max}" for name, (min, max) in sorted(box.items())
    )
    rule = designspaceLib.RuleDescriptor()
    rule.name = "BRACKET." + description
    rule.conditionSets.append(
        [
            {"name": axis_name, "minimum": range_min, "maximum": range_max}
            for axis_name, (range_min, range_max) in box.items()
        ]
    )
    rule.subs = list(mapping.items())
    return rule


def _expand_kerning_to_brackets(
    glyph_name: str, ufo_glyph_name: str, ufo_font: Any
) -> None:
    """Ensures that bracket glyphs inherit their parents' kerning."""

    for group, names in ufo_font.groups.items():
        if not group.startswith(("public.kern1.", "public.kern2.")):
            continue
        name_set = set(names)
        if glyph_name in name_set and ufo_glyph_name not in name_set:
            names.append(ufo_glyph_name)

    bracket_kerning = {}
    for (first, second), value in ufo_font.kerning.items():
        first_match = first == glyph_name
        second_match = second == glyph_name
        if first_match and second_match:
            bracket_kerning[(ufo_glyph_name, ufo_glyph_name)] = value
        elif first_match:
            bracket_kerning[(ufo_glyph_name, second)] = value
        elif second_match:
            bracket_kerning[(first, ufo_glyph_name)] = value
    ufo_font.kerning.update(bracket_kerning)



#
# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
from textwrap import dedent

from glyphsLib import to_glyphs, to_ufos, classes, to_designspace
from glyphsLib.builder.features import _build_public_opentype_categories

from fontTools.designspaceLib import DesignSpaceDocument
import pytest


def roundtrip(ufo, tmpdir, ufo_module):
    font = to_glyphs([ufo], minimize_ufo_diffs=True)
    filename = os.path.join(str(tmpdir), "font.glyphs")
    font.save(filename)
    font = classes.GSFont(filename)
    (ufo,) = to_ufos(font, ufo_module=ufo_module)
    return font, ufo


def test_blank(tmpdir, ufo_module):
    ufo = ufo_module.Font()

    font, rtufo = roundtrip(ufo, tmpdir, ufo_module)

    assert not font.features
    assert not font.featurePrefixes
    assert not rtufo.features.text


def test_comment(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    ufo.features.text = dedent(
        """\
        # Test
        # Lol
    """
    )

    font, rtufo = roundtrip(ufo, tmpdir, ufo_module)

    assert not font.features
    assert len(font.featurePrefixes) == 1
    fp = font.featurePrefixes[0]
    assert fp.code.strip() == ufo.features.text.strip()
    assert not fp.automatic

    assert rtufo.features.text == ufo.features.text


def test_languagesystems(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    # The sample has messed-up spacing because there was a problem with that
    ufo.features.text = dedent(
        """\
        # Blah
          languagesystem DFLT dflt; #Default
        languagesystem latn dflt;\t# Latin
        \tlanguagesystem arab URD; #\tUrdu
    """
    )

    font, rtufo = roundtrip(ufo, tmpdir, ufo_module)

    assert not font.features
    assert len(font.featurePrefixes) == 1
    fp = font.featurePrefixes[0]
    assert fp.code == ufo.features.text[:-1]  # Except newline
    assert not fp.automatic

    assert rtufo.features.text == ufo.features.text


def test_classes(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    # FIXME: (jany) no whitespace is preserved in this section
    ufo.features.text = dedent(
        """\
        @lc = [ a b
        ];

        @UC = [ A B
        ];

        @all = [ @lc @UC zero one
        ];

        @more = [ dot @UC colon @lc paren
        ];
    """
    )

    font, rtufo = roundtrip(ufo, tmpdir, ufo_module)

    assert len(font.classes) == 4
    assert font.classes["lc"].code == "a b"
    assert font.classes["UC"].code == "A B"
    assert font.classes["all"].code == "@lc @UC zero one"
    assert font.classes["more"].code == "dot @UC colon @lc paren"

    assert rtufo.features.text == ufo.features.text


def test_class_synonym(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    ufo.features.text = dedent(
        """\
        @lc = [ a b
        ];

        @lower = @lc;
    """
    )

    font, rtufo = roundtrip(ufo, tmpdir, ufo_module)

    assert len(font.classes) == 2
    assert font.classes["lc"].code == "a b"
    assert font.classes["lower"].code == "@lc"

    # FIXME: (jany) should roundtrip
    assert rtufo.features.text == dedent(
        """\
        @lc = [ a b
        ];

        @lower = [ @lc
        ];
    """
    )


def test_feature_names(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    ufo.features.text = dedent(
        """\
        feature ss01 {
        featureNames {
          name "Alternate g";
        };
        # automatic
        sub g by g.ss01;

        } ss01;
    """
    )

    font, rtufo = roundtrip(ufo, tmpdir, ufo_module)

    # Check code in Glyphs font
    gs_feature = font.features[0]
    assert gs_feature.automatic
    assert gs_feature.code.strip() == "sub g by g.ss01;"
    assert gs_feature.notes.strip() == "Name: Alternate g"

    assert rtufo.features.text == dedent(
        """\
        feature ss01 {
        featureNames {
          name "Alternate g";
        };
        # automatic
        sub g by g.ss01;

        } ss01;
    """
    )


def test_feature_names_notes(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    ufo.features.text = dedent(
        """\
        feature ss01 {
        # notes:
        # foo
        featureNames {
          name "Alternate g";
        };
        # automatic
        sub g by g.ss01;

        } ss01;
    """
    )

    font, rtufo = roundtrip(ufo, tmpdir, ufo_module)

    # Check code in Glyphs font
    gs_feature = font.features[0]
    assert gs_feature.automatic
    assert gs_feature.code.strip() == "sub g by g.ss01;"
    assert gs_feature.notes.strip() == "Name: Alternate g\nfoo"

    assert rtufo.features.text == dedent(
        """\
        feature ss01 {
        # notes:
        # foo
        featureNames {
          name "Alternate g";
        };
        # automatic
        sub g by g.ss01;

        } ss01;
    """
    )


def test_feature_names_full(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    ufo.features.text = dedent(
        """\
        feature ss01 {
        featureNames {
          name 1 "Alternate g";
        };
        # automatic
        sub g by g.ss01;

        } ss01;
    """
    )

    font, rtufo = roundtrip(ufo, tmpdir, ufo_module)

    # Check code in Glyphs font
    gs_feature = font.features[0]
    assert gs_feature.automatic
    assert gs_feature.code.strip() == "sub g by g.ss01;"
    assert gs_feature.notes.strip() == dedent(
        """\
        featureNames {
            name 1 "Alternate g";
        };"""
    )

    assert rtufo.features.text == dedent(
        """\
        feature ss01 {
        featureNames {
            name 1 "Alternate g";
        };
        # automatic
        sub g by g.ss01;

        } ss01;
    """
    )


def test_feature_names_multi(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    ufo.features.text = dedent(
        """\
        feature ss01 {
        featureNames {
          name "Alternate g";
          name 1 "Alternate g";
        };
        # automatic
        sub g by g.ss01;

        } ss01;
    """
    )

    font, rtufo = roundtrip(ufo, tmpdir, ufo_module)

    # Check code in Glyphs font
    gs_feature = font.features[0]
    assert gs_feature.automatic
    assert gs_feature.code.strip() == "sub g by g.ss01;"
    assert gs_feature.notes.strip() == dedent(
        """\
        featureNames {
            name "Alternate g";
            name 1 "Alternate g";
        };"""
    )

    assert rtufo.features.text == dedent(
        """\
        feature ss01 {
        featureNames {
            name "Alternate g";
            name 1 "Alternate g";
        };
        # automatic
        sub g by g.ss01;

        } ss01;
    """
    )


def test_include(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    ufo.features.text = dedent(
        """\
        include(../family.fea);
        # Blah
        include(../fractions.fea);
    """
    )

    font, rtufo = roundtrip(ufo, tmpdir, ufo_module)

    assert len(font.featurePrefixes) == 1
    assert font.featurePrefixes[0].code.strip() == ufo.features.text.strip()

    assert rtufo.features.text == ufo.features.text


def test_include_no_semicolon(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    ufo.features.text = dedent(
        """\
        include(../family.fea)
    """
    )

    font, rtufo = roundtrip(ufo, tmpdir, ufo_module)

    assert len(font.featurePrefixes) == 1
    assert font.featurePrefixes[0].code.strip() == ufo.features.text.strip()

    assert rtufo.features.text == ufo.features.text


def test_to_glyphs_expand_includes(tmp_path, ufo_module):
    ufo = ufo_module.Font()
    ufo.features.text = dedent(
        """\
        include(family.fea);
        """
    )
    ufo.save(str(tmp_path / "font.ufo"))

    included_path = tmp_path / "family.fea"
    included_path.write_text("# hello from family.fea")
    assert included_path.exists()

    font = to_glyphs([ufo], minimize_ufo_diffs=True, expand_includes=True)

    assert len(font.featurePrefixes) == 1
    assert font.featurePrefixes[0].code.strip() == "# hello from family.fea"


def test_to_ufos_expand_includes(tmp_path, ufo_module):
    font = classes.GSFont()
    font.masters.append(classes.GSFontMaster())

    feature_prefix = classes.GSFeaturePrefix()
    feature_prefix.name = "include"
    feature_prefix.code = dedent(
        """\
        include(family.fea);
        """
    )
    font.featurePrefixes.append(feature_prefix)

    font.filepath = str(tmp_path / "font.glyphs")
    font.save()

    included_path = tmp_path / "family.fea"
    included_path.write_text("# hello from family.fea")
    assert included_path.exists()

    (ufo,) = to_ufos(font, ufo_module=ufo_module, expand_includes=True)

    assert ufo.features.text == ("# Prefix: include\n# hello from family.fea")


def test_standalone_lookup(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    # FIXME: (jany) does not preserve whitespace before and after
    ufo.features.text = dedent(
        """\
        # start of default rules that are applied under all language systems.
        lookup HAS_I {
          sub f f i by f_f_i;
            sub f i by f_i;
        } HAS_I;
    """
    )

    font, rtufo = roundtrip(ufo, tmpdir, ufo_module)

    assert len(font.featurePrefixes) == 1
    assert font.featurePrefixes[0].code.strip() == ufo.features.text.strip()

    assert rtufo.features.text == ufo.features.text


def test_feature(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    # This sample is straight from the documentation at
    # http://www.adobe.com/devnet/opentype/afdko/topic_feature_file_syntax.html
    # FIXME: (jany) does not preserve whitespace before and after
    ufo.features.text = dedent(
        """\
        feature liga {
      # start of default rules that are applied under all language systems.
                lookup HAS_I {
                 sub f f i by f_f_i;
                 sub f i by f_i;
                } HAS_I;

                lookup NO_I {
                 sub f f l by f_f_l;
                 sub f f by f_f;
                } NO_I;

      # end of default rules that are applied under all language systems.

            script latn;
               language dflt;
      # default lookup for latn included under all languages for the latn script

                  sub f l by f_l;
               language DEU;
      # default lookups included under the DEU language..
                  sub s s by germandbls;   # This is also included.
               language TRK exclude_dflt;   # default lookups are excluded.
                lookup NO_I;             # Only this lookup is included
                                         # under the TRK language

            script cyrl;
               language SRB;
                  sub c t by c_t; # this rule will apply only under
                                  # script cyrl language SRB.
      } liga;
    """
    )

    font, rtufo = roundtrip(ufo, tmpdir, ufo_module)

    assert len(font.features) == 1
    # Strip "feature liga {} liga;"
    code = ufo.features.text.splitlines()[1:-1]
    assert font.features[0].code.strip() == "\n".join(code)

    assert rtufo.features.text.strip() == ufo.features.text.strip()


def test_different_features_in_different_UFOS(tmpdir, ufo_module):
    # If the input UFOs have different features, Glyphs cannot model the
    # differences easily.
    #
    # TODO: (jany) A complex solution would be to put all the features that we
    # find across all the UFOS into the GSFont's features, and then add custom
    # parameters "Replace Features" and "Remove features" to the GSFontMasters
    # of the UFOs that didn't have the feature originally.
    #
    # What is done now: if feature files differ between the input UFOs, the
    # original text of each UFO's feature is stored in userData, and a single
    # GSFeaturePrefix is created just to warn the user that features were not
    # imported because of differences.
    ufo1 = ufo_module.Font()
    ufo1.features.text = dedent(
        """\
        include('../family.fea');
    """
    )
    ufo2 = ufo_module.Font()
    ufo2.features.text = dedent(
        """\
        include('../family.fea');

        feature ss03 {
            sub c by c.ss03;
        } ss03;
    """
    )

    font = to_glyphs([ufo1, ufo2], minimize_ufo_diffs=True)
    filename = os.path.join(str(tmpdir), "font.glyphs")
    font.save(filename)
    font = classes.GSFont(filename)
    ufo1rt, ufo2rt = to_ufos(font, ufo_module=ufo_module)

    assert len(font.features) == 0
    assert len(font.featurePrefixes) == 1
    assert font.featurePrefixes[0].code == dedent(
        """\
        # Do not use Glyphs to edit features.
        #
        # This Glyphs file was made from several UFOs that had different
        # features. As a result, the features are not editable in Glyphs and
        # the original features will be restored when you go back to UFOs.
    """
    )

    assert ufo1rt.features.text == ufo1.features.text
    assert ufo2rt.features.text == ufo2.features.text


def test_roundtrip_disabled_feature(ufo_module):
    font = to_glyphs([ufo_module.Font()])
    feature = classes.GSFeature(name="ccmp")
    feature.code = dedent(
        """\
        sub a by a.ss03;
        sub b by b.ss03;
        sub c by c.ss03;
    """
    )
    feature.disabled = True
    font.features.append(feature)

    (ufo,) = to_ufos(font, ufo_module=ufo_module)
    assert ufo.features.text == dedent(
        """\
        feature ccmp {
        # disabled
        #sub a by a.ss03;
        #sub b by b.ss03;
        #sub c by c.ss03;
        } ccmp;
    """
    )

    font_r = to_glyphs([ufo])
    assert len(font_r.features) == 1
    feature_r = font_r.features[0]
    assert feature_r.name == "ccmp"
    assert feature_r.code == feature.code
    assert feature_r.disabled is True

    font_rr = to_glyphs(to_ufos(font_r, ufo_module=ufo_module))
    assert len(font_rr.features) == 1
    feature_rr = font_rr.features[0]
    assert feature_rr.name == "ccmp"
    assert feature_rr.code == feature.code
    assert feature_rr.disabled is True


def test_roundtrip_automatic_feature(ufo_module):
    font = to_glyphs([ufo_module.Font()])
    feature = classes.GSFeature(name="ccmp")
    feature.code = "sub c by c.ss03;"
    feature.automatic = True
    font.features.append(feature)

    (ufo,) = to_ufos(font, ufo_module=ufo_module)
    assert ufo.features.text == dedent(
        """\
        feature ccmp {
        # automatic
        sub c by c.ss03;
        } ccmp;
    """
    )

    font_r = to_glyphs([ufo])
    assert len(font_r.features) == 1
    feature_r = font_r.features[0]
    assert feature_r.name == "ccmp"
    assert feature_r.code == "sub c by c.ss03;"
    assert feature_r.automatic is True


def test_roundtrip_feature_prefix_with_only_a_comment(ufo_module):
    font = to_glyphs([ufo_module.Font()])
    prefix = classes.GSFeaturePrefix(name="include")
    # Contents: just a comment
    prefix.code = "#include(../family.fea)"
    font.featurePrefixes.append(prefix)

    (ufo,) = to_ufos(font, ufo_module=ufo_module)

    assert ufo.features.text == dedent(
        """\
        # Prefix: include
        #include(../family.fea)
    """
    )

    font_r = to_glyphs([ufo])
    assert len(font_r.featurePrefixes) == 1
    prefix_r = font_r.featurePrefixes[0]
    assert prefix_r.name == "include"
    assert prefix_r.code == "#include(../family.fea)"


def test_drop_disabled_class(ufo_module):
    font = to_glyphs([ufo_module.Font()])
    class_ = classes.GSClass(name="Class1", code="a b")
    class_.disabled = True
    font.classes.append(class_)

    class_ = classes.GSClass(name="Class2", code="c d")
    font.classes.append(class_)

    (ufo,) = to_ufos(font, ufo_module=ufo_module, minimal=True)
    assert ufo.features.text == dedent(
        """\
        @Class2 = [ c d
        ];
    """
    )


def test_drop_disabled_prefix(ufo_module):
    font = to_glyphs([ufo_module.Font()])
    prefix = classes.GSFeaturePrefix(name="Prefix1", code="# test 1")
    prefix.disabled = True
    font.featurePrefixes.append(prefix)

    prefix = classes.GSFeaturePrefix(name="Prefix2", code="# test 2")
    font.featurePrefixes.append(prefix)

    (ufo,) = to_ufos(font, ufo_module=ufo_module, minimal=True)
    assert ufo.features.text == dedent(
        """\
        # Prefix: Prefix2
        # test 2
    """
    )


def test_drop_disabled_feature(ufo_module):
    font = to_glyphs([ufo_module.Font()])
    feature = classes.GSFeature(name="ccmp", code="sub a by a.ccmp1 a.ccmp2;")
    feature.disabled = True
    font.features.append(feature)

    feature = classes.GSFeature(name="liga", code="sub f i by f_i;")
    font.features.append(feature)

    (ufo,) = to_ufos(font, ufo_module=ufo_module, minimal=True)
    assert ufo.features.text == dedent(
        """\
        feature liga {
        sub f i by f_i;
        } liga;
    """
    )


@pytest.fixture
def ufo_with_GDEF(ufo_module):
    ufo = ufo_module.Font()
    gdef = dedent(
        """\
        table GDEF {
        GlyphClassDef
            [A], # Base
            , # Liga
            [dieresiscomb], # Mark
            ;
        } GDEF;
        """
    )
    ufo.features.text = gdef
    return ufo, gdef, ufo_module


def test_roundtrip_existing_GDEF(tmpdir, ufo_with_GDEF):
    """Test that an existing GDEF table in UFO is preserved unchanged and
    no extra GDEF table is generated upon roundtripping to UFO when
    `generate_GDEF` is False.
    """
    ufo, gdef, ufo_module = ufo_with_GDEF
    font = to_glyphs([ufo])
    filename = os.path.join(str(tmpdir), "font.glyphs")
    font.save(filename)
    font = classes.GSFont(filename)
    (rtufo,) = to_ufos(font, generate_GDEF=False, ufo_module=ufo_module)

    assert rtufo.features.text == gdef


def test_groups_remain_at_top(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    ufo.newGlyph("zero")
    ufo.newGlyph("zero.alt")
    fea_example = dedent(
        """\
        @FIG_DFLT = [zero];
        @FIG_ALT = [zero.alt];

        lookup pnum_text {
            sub @FIG_DFLT by @FIG_ALT;
        } pnum_text;

        feature pnum {
            lookup pnum_text;
        } pnum;
        """
    )
    ufo.features.text = fea_example

    font = to_glyphs([ufo], minimize_ufo_diffs=True)
    filename = os.path.join(str(tmpdir), "font.glyphs")
    font.save(filename)
    font = classes.GSFont(filename)
    (rtufo,) = to_ufos(font, ufo_module=ufo_module)

    fea_rt = rtufo.features.text
    assert fea_rt.index("@FIG_DFLT") < fea_rt.index("lookup") < fea_rt.index("feature")


def test_roundtrip_empty_feature(ufo_module):
    # https://github.com/googlefonts/glyphsLib/issues/562
    font = to_glyphs([ufo_module.Font()])
    feature = classes.GSFeature(name="dlig")
    feature.code = ""
    font.features.append(feature)

    (ufo,) = to_ufos(font, ufo_module=ufo_module)
    assert ufo.features.text == dedent(
        """\
        feature dlig {

        } dlig;
    """
    )

    font_r = to_glyphs([ufo])
    assert len(font_r.features) == 1
    feature_r = font_r.features[0]
    assert feature_r.name == "dlig"
    assert feature_r.code == ""


def test_build_GDEF_incomplete_glyphOrder():
    import ufoLib2

    font = ufoLib2.Font()
    font.lib["public.glyphOrder"] = ["b", "c"]
    for name in ("d", "c", "b", "a"):
        glyph = font.newGlyph(name)
        glyph.appendAnchor({"name": "top", "x": 0, "y": 0})

    categories = _build_public_opentype_categories(font)
    assert categories == {"a": "base", "b": "base", "c": "base", "d": "base"}


def test_comments_in_classes(ufo_module):
    filename = os.path.join(os.path.dirname(__file__), "../data/CommentedClass.glyphs")
    font = classes.GSFont(filename)
    (ufo,) = to_ufos(font)
    assert ufo.features.text == dedent(
        """\
            @Test = [ A
            # B
            ];
"""
    )


def test_mark_class_used_as_glyph_class(tmpdir, ufo_module):
    ufo = ufo_module.Font()
    ufo.features.text = dedent(
        """\
        markClass CombBreve_Macron <anchor -700 1000> @_U;
        markClass CombGraphemeJoiner <anchor 0 0> @_U;
        @c_u_diacs = @_U;
        """
    )

    font = to_glyphs([ufo], minimize_ufo_diffs=True)

    assert (
        font.featurePrefixes[0].code
        == dedent(
            """\
            markClass CombBreve_Macron <anchor -700 1000> @_U;
            markClass CombGraphemeJoiner <anchor 0 0> @_U;
            """
        ).strip()
    )
    assert font.classes[0].name == "c_u_diacs"
    assert font.classes[0].code == "@_U"

    (rtufo,) = to_ufos(font, ufo_module=ufo_module)

    # After roundtrip glyph classes are always placed before the mark classes
    # hence the following assertion would fail...
    # https://github.com/googlefonts/glyphsLib/issues/694#issuecomment-1117204523
    # assert rtufo.features.text == ufo.features.text


def test_automatic_added_to_manual_kern(tmpdir, ufo_module):
    """Test that when a Glyphs file has a manual kern feature,
    automatic markers are added so that the source kerning also
    gets applied.
    """
    font = classes.GSFont()
    font.masters.append(classes.GSFontMaster())

    (ufo,) = to_ufos(font)

    assert "# Automatic Code" not in ufo.features.text

    kern = classes.GSFeature(name="kern", code="pos a b 100;")
    font.features.append(kern)
    (ufo,) = to_ufos(font)

    assert "# Automatic Code" in ufo.features.text

    designspace = to_designspace(font, ufo_module=ufo_module)
    path = str(tmpdir / "test.designspace")
    designspace.write(path)
    for source in designspace.sources:
        source.font.save(str(tmpdir / source.filename))

    designspace2 = DesignSpaceDocument.fromfile(path)
    font2 = to_glyphs(designspace2, ufo_module=ufo_module)

    assert len([f for f in font2.features if f.name == "kern"]) == 1
